from __future__ import annotations

import typing
from collections import Counter, abc, defaultdict
from itertools import product

from stk._internal.building_block import BuildingBlock
from stk._internal.optimizers.null import NullOptimizer
from stk._internal.optimizers.optimizer import Optimizer
from stk._internal.reaction_factories.dative_reaction_factory import (
    DativeReactionFactory,
)
from stk._internal.reaction_factories.generic_reaction_factory import (
    GenericReactionFactory,
)
from stk._internal.reaction_factories.reaction_factory import (
    ReactionFactory,
)
from stk._internal.topology_graphs.edge import Edge
from stk._internal.topology_graphs.topology_graph.topology_graph import (
    TopologyGraph,
)
from stk._internal.topology_graphs.vertex import Vertex
from stk._internal.utilities.utilities import OneOrMany

from .vertices import (
    BiDentateLigandVertex,
    MetalVertex,
    MonoDentateLigandVertex,
    UnaligningVertex,
)

_LigandVertex = typing.Union[
    UnaligningVertex,
    MonoDentateLigandVertex,
    BiDentateLigandVertex,
]


class MetalComplex(TopologyGraph):
    """
    Represents a metal complex topology graph.

    Notes:

        *Subclass Implementation*

        Each subclass needs to define the attributes,
        :attr:`_metal_vertex_prototypes` and
        :attr:`_ligand_vertex_prototypes`, which are :class:`tuple` of
        :class:`.Vertex` instances.

    Examples:

        *Subclass Implementation*

        The source code of the subclasses, listed in
        :mod:`~.metal_complex.metal_complex`, can serve as good
        examples.

        *Basic Construction*

        For most :class:`.MetalComplex` topology graphs, we first
        need to define a metal :class:`.BuildingBlock`, consisting of
        1 atom and multiple functional groups

        .. testcode:: basic-construction

            import stk

            metal = stk.BuildingBlock(
                smiles='[Fe+2]',
                functional_groups=(
                    stk.SingleAtom(stk.Fe(0, charge=2))
                    for i in range(6)
                ),
                position_matrix=[[0, 0, 0]],
            )

        We also need to define an organic ligand
        :class:`.BuildingBlock`

        .. testcode:: basic-construction

            # Define an organic linker with two functional groups.
            bidentate = stk.BuildingBlock(
                smiles='C=NC/C=N/Br',
                functional_groups=[
                    stk.SmartsFunctionalGroupFactory(
                        smarts='[#6]~[#7X2]~[#35]',
                        bonders=(1, ),
                        deleters=(),
                    ),
                    stk.SmartsFunctionalGroupFactory(
                        smarts='[#6]~[#7X2]~[#6]',
                        bonders=(1, ),
                        deleters=(),
                    ),
                ],
            )

        .. moldoc::

            import moldoc.molecule as molecule
            import stk

            bidentate = stk.BuildingBlock(
                smiles='C=NC/C=N/Br',
                functional_groups=[
                    stk.SmartsFunctionalGroupFactory(
                        smarts='[#6]~[#7X2]~[#35]',
                        bonders=(1, ),
                        deleters=(),
                    ),
                    stk.SmartsFunctionalGroupFactory(
                        smarts='[#6]~[#7X2]~[#6]',
                        bonders=(1, ),
                        deleters=(),
                    ),
                ],
            )

            moldoc_display_molecule = molecule.Molecule(
                atoms=(
                    molecule.Atom(
                        atomic_number=atom.get_atomic_number(),
                        position=position,
                    ) for atom, position in zip(
                        bidentate.get_atoms(),
                        bidentate.get_position_matrix(),
                    )
                ),
                bonds=(
                    molecule.Bond(
                        atom1_id=bond.get_atom1().get_id(),
                        atom2_id=bond.get_atom2().get_id(),
                        order=bond.get_order(),
                    ) for bond in bidentate.get_bonds()
                ),
            )


        Finally, we can create the :class:`.MetalComplex`.

        .. testcode:: basic-construction

            complex = stk.ConstructedMolecule(
                topology_graph=stk.metal_complex.OctahedralLambda(
                    metals=metal,
                    ligands=bidentate,
                )
            )

        .. moldoc::

            import moldoc.molecule as molecule
            import stk

            metal = stk.BuildingBlock(
                smiles='[Fe+2]',
                functional_groups=(
                    stk.SingleAtom(stk.Fe(0, charge=2))
                    for i in range(6)
                ),
                position_matrix=[[0, 0, 0]],
            )

            bidentate = stk.BuildingBlock(
                smiles='C=NC/C=N/Br',
                functional_groups=[
                    stk.SmartsFunctionalGroupFactory(
                        smarts='[#6]~[#7X2]~[#35]',
                        bonders=(1, ),
                        deleters=(),
                    ),
                    stk.SmartsFunctionalGroupFactory(
                        smarts='[#6]~[#7X2]~[#6]',
                        bonders=(1, ),
                        deleters=(),
                    ),
                ],
            )

            complex = stk.ConstructedMolecule(
                topology_graph=stk.metal_complex.OctahedralLambda(
                    metals=metal,
                    ligands=bidentate,
                )
            )

            moldoc_display_molecule = molecule.Molecule(
                atoms=(
                    molecule.Atom(
                        atomic_number=atom.get_atomic_number(),
                        position=position,
                    ) for atom, position in zip(
                        complex.get_atoms(),
                        complex.get_position_matrix(),
                    )
                ),
                bonds=(
                    molecule.Bond(
                        atom1_id=bond.get_atom1().get_id(),
                        atom2_id=bond.get_atom2().get_id(),
                        order=(
                            bond.get_order()
                            if bond.get_order() != 9
                            else 1
                        ),
                    ) for bond in complex.get_bonds()
                ),
            )

        *Suggested Optimization*

        For :class:`.MetalComplex` topologies, it is recommend to use
        the :class:`.MCHammer` optimizer.

        .. testcode:: suggested-optimization

            import stk

            metal = stk.BuildingBlock(
                smiles='[Fe+2]',
                functional_groups=(
                    stk.SingleAtom(stk.Fe(0, charge=2))
                    for i in range(6)
                ),
                position_matrix=[[0, 0, 0]],
            )

            bidentate = stk.BuildingBlock(
                smiles='C=NC/C=N/Br',
                functional_groups=[
                    stk.SmartsFunctionalGroupFactory(
                        smarts='[#6]~[#7X2]~[#35]',
                        bonders=(1, ),
                        deleters=(),
                    ),
                    stk.SmartsFunctionalGroupFactory(
                        smarts='[#6]~[#7X2]~[#6]',
                        bonders=(1, ),
                        deleters=(),
                    ),
                ]
            )

            complex = stk.ConstructedMolecule(
                topology_graph=stk.metal_complex.OctahedralLambda(
                    metals=metal,
                    ligands=bidentate,
                    optimizer=stk.MCHammer(),
                ),
            )

        .. moldoc::

            import moldoc.molecule as molecule
            import stk

            metal = stk.BuildingBlock(
                smiles='[Fe+2]',
                functional_groups=(
                    stk.SingleAtom(stk.Fe(0, charge=2))
                    for i in range(6)
                ),
                position_matrix=[[0, 0, 0]],
            )

            bidentate = stk.BuildingBlock(
                smiles='C=NC/C=N/Br',
                functional_groups=[
                    stk.SmartsFunctionalGroupFactory(
                        smarts='[#6]~[#7X2]~[#35]',
                        bonders=(1, ),
                        deleters=(),
                    ),
                    stk.SmartsFunctionalGroupFactory(
                        smarts='[#6]~[#7X2]~[#6]',
                        bonders=(1, ),
                        deleters=(),
                    ),
                ]
            )

            complex = stk.ConstructedMolecule(
                topology_graph=stk.metal_complex.OctahedralLambda(
                    metals=metal,
                    ligands=bidentate,
                    optimizer=stk.MCHammer(),
                ),
            )

            moldoc_display_molecule = molecule.Molecule(
                atoms=(
                    molecule.Atom(
                        atomic_number=atom.get_atomic_number(),
                        position=position,
                    ) for atom, position in zip(
                        complex.get_atoms(),
                        complex.get_position_matrix(),
                    )
                ),
                bonds=(
                    molecule.Bond(
                        atom1_id=bond.get_atom1().get_id(),
                        atom2_id=bond.get_atom2().get_id(),
                        order=(
                            bond.get_order()
                            if bond.get_order() != 9
                            else 1
                        ),
                    ) for bond in complex.get_bonds()
                ),
            )

        *Construction with Multiple Metals & Ligands*

        When multiple metals or ligands are used, the `metals` and
        `ligands` parameters accept values of type :class:`dict`, which
        specify the exact vertex each metal or ligand needs to be
        placed on.

        .. testcode:: construction-with-multiple-metals-and-ligands

            import stk

            metal = stk.BuildingBlock(
                smiles='[Fe+2]',
                functional_groups=(
                    stk.SingleAtom(stk.Fe(0, charge=2))
                    for i in range(6)
                ),
                position_matrix=[[0, 0, 0]],
            )

            bidentate1 = stk.BuildingBlock(
                smiles='C=NC/C=N/Br',
                functional_groups=[
                    stk.SmartsFunctionalGroupFactory(
                        smarts='[#6]~[#7X2]~[#35]',
                        bonders=(1, ),
                        deleters=(),
                    ),
                    stk.SmartsFunctionalGroupFactory(
                        smarts='[#6]~[#7X2]~[#6]',
                        bonders=(1, ),
                        deleters=(),
                    ),
                ],
            )

            # Define a second organic linker with two functional
            # groups.
            bidentate2 = stk.BuildingBlock(
                smiles='C=NC(C)(C)/C(C)=N/Br',
                functional_groups=[
                    stk.SmartsFunctionalGroupFactory(
                        smarts='[#6]~[#7X2]~[#35]',
                        bonders=(1, ),
                        deleters=(),
                    ),
                    stk.SmartsFunctionalGroupFactory(
                        smarts='[#6]~[#7X2]~[#6]',
                        bonders=(1, ),
                        deleters=(),
                    ),
                ],
            )

            # Build heteroleptic complex.
            complex = stk.ConstructedMolecule(
                topology_graph=stk.metal_complex.OctahedralLambda(
                    metals=metal,
                    ligands={
                        bidentate1: (0, 1),
                        bidentate2: (2, ),
                    },
                ),
            )

        However, if each ligand is has a different number of
        functional groups, they can be provided together in a
        :class:`tuple`.

        Note that the valid vertex identifiers depend on the exact
        metal complex you are using. These are detailed in the
        docstring for that specific metal vertex topology graph.

        *Unsubstituted Metal Complexes*

        Some metal complex topologies represent metal complexes with
        unsubstituted metal sites. For example,
        :class:`.BidentateSquarePlanar` has all sites substituted and
        :class:`.CisProtectedSquarePlanar` is the equivalent metal
        complex with some unsubstituted sites

        .. testcode:: leaving-unsubstituted-sites

            import stk

            pd = stk.BuildingBlock(
                smiles='[Pd+2]',
                functional_groups=(
                    stk.SingleAtom(stk.Pd(0, charge=2))
                    for i in range(4)
                ),
                position_matrix=[[0, 0, 0]],
            )

            # Define a bidentate ligand with two functional groups.
            bidentate_ligand = stk.BuildingBlock(
                smiles='NCCN',
                functional_groups=[
                    stk.SmartsFunctionalGroupFactory(
                        smarts='[#7]~[#6]',
                        bonders=(0, ),
                        deleters=(),
                    ),
                ],
            )

            # Construct a cis-protected square planar metal complex.
            complex = stk.ConstructedMolecule(
                topology_graph=stk.metal_complex.CisProtectedSquarePlanar(
                    metals=pd,
                    ligands=bidentate_ligand,
                ),
            )

        .. moldoc::

            import moldoc.molecule as molecule
            import stk

            pd = stk.BuildingBlock(
                smiles='[Pd+2]',
                functional_groups=(
                    stk.SingleAtom(stk.Pd(0, charge=2))
                    for i in range(4)
                ),
                position_matrix=[[0, 0, 0]],
            )

            # Define a bidentate ligand with two functional groups.
            bidentate_ligand = stk.BuildingBlock(
                smiles='NCCN',
                functional_groups=[
                    stk.SmartsFunctionalGroupFactory(
                        smarts='[#7]~[#6]',
                        bonders=(0, ),
                        deleters=(),
                    ),
                ],
            )

            # Construct a cis-protected square planar metal complex.
            complex = stk.ConstructedMolecule(
                topology_graph=stk.metal_complex.CisProtectedSquarePlanar(
                    metals=pd,
                    ligands=bidentate_ligand,
                ),
            )

            moldoc_display_molecule = molecule.Molecule(
                atoms=(
                    molecule.Atom(
                        atomic_number=atom.get_atomic_number(),
                        position=position,
                    ) for atom, position in zip(
                        complex.get_atoms(),
                        complex.get_position_matrix(),
                    )
                ),
                bonds=(
                    molecule.Bond(
                        atom1_id=bond.get_atom1().get_id(),
                        atom2_id=bond.get_atom2().get_id(),
                        order=(
                            bond.get_order()
                            if bond.get_order() != 9
                            else 1
                        ),
                    ) for bond in complex.get_bonds()
                ),
            )

    """

    _edge_prototypes: typing.ClassVar[tuple[Edge, ...]]
    _metal_vertex_prototypes: typing.ClassVar[tuple[MetalVertex, ...]]
    _ligand_vertex_prototypes: typing.ClassVar[tuple[_LigandVertex, ...],]
    _metal_vertices_of_degree: typing.ClassVar[abc.Mapping[int, set[int]],]
    _ligand_vertices_of_degree: typing.ClassVar[abc.Mapping[int, set[int]],]
    _vertex_degrees: typing.ClassVar[abc.Mapping[int, int]]

    def __init_subclass__(cls, **kwargs: typing.Any) -> None:
        cls._vertex_degrees = Counter(
            vertex_id
            for edge in cls._edge_prototypes
            for vertex_id in edge.get_vertex_ids()
        )
        cls._metal_vertices_of_degree = defaultdict(set)
        for metal_vertex in cls._metal_vertex_prototypes:
            degree = cls._vertex_degrees[metal_vertex.get_id()]
            cls._metal_vertices_of_degree[degree].add(
                metal_vertex.get_id(),
            )

        cls._ligand_vertices_of_degree = defaultdict(set)
        for ligand_vertex in cls._ligand_vertex_prototypes:
            degree = cls._vertex_degrees[ligand_vertex.get_id()]
            cls._ligand_vertices_of_degree[degree].add(
                ligand_vertex.get_id(),
            )

    def __init__(
        self,
        metals: typing.Union[
            BuildingBlock,
            tuple[BuildingBlock, ...],
            dict[BuildingBlock, tuple[int, ...]],
        ],
        ligands: typing.Union[
            BuildingBlock,
            tuple[BuildingBlock, ...],
            dict[BuildingBlock, tuple[int, ...]],
        ],
        reaction_factory: typing.Optional[ReactionFactory] = None,
        num_processes: int = 1,
        optimizer: Optimizer = NullOptimizer(),
    ) -> None:
        """
        Initialize a :class:`.MetalComplex`.

        Parameters:

            metals:
                Can be a :class:`dict` which maps the
                :class:`.BuildingBlock` instances to the indices of the
                vertices in :attr:`_metal_vertex_prototypes` it should
                be placed on.

                If each :class:`.BuildingBlock` has a different number
                of functional groups, they can be supplied together in
                a :class:`tuple`.

                Can also be a :class:`.BuildingBlock` instance, which
                should be placed on all
                :attr:`_metal_vertex_prototypes` on the topology graph.

            ligands:
                Can be a :class:`dict` which maps the
                :class:`.BuildingBlock` instances to the indices of the
                vertices in :attr:`_ligand_vertex_prototypes` it should
                be placed on.

                If each :class:`.BuildingBlock` has a different number
                of functional groups, they can be supplied together in
                a :class:`tuple`.

                Can also be a :class:`.BuildingBlock` instance, which
                should be placed on all
                :attr:`_ligand_vertex_prototypes` on the topology
                graph.

            reaction_factory:
                The reaction factory to use for creating bonds between
                building blocks. If ``None``, a
                :class:`.DativeReactionFactory` is used, which produces
                only dative bonds in any reactions done by this
                topology construction.

            num_processes:
                The number of parallel processes to create during
                :meth:`construct`.

            optimizer:
                Used to optimize the structure of the constructed
                molecule.

        """

        building_block_vertices = self._normalize_metals(metals)
        building_block_vertices.update(
            (building_block, vertices)
            for building_block, vertices in self._normalize_ligands(
                ligands
            ).items()
        )

        # By default, assign a dative bond order to available
        # functional groups.
        if reaction_factory is None:
            metal_functional_groups = set(
                type(functional_group)
                for metal in self._normalize_metals(metals)
                for functional_group in metal.get_functional_groups()
            )
            ligand_functional_groups = set(
                type(functional_group)
                for ligand in self._normalize_ligands(ligands)
                for functional_group in ligand.get_functional_groups()
            )
            functional_group_pairs = product(
                metal_functional_groups,
                ligand_functional_groups,
            )
            reaction_factory = DativeReactionFactory(
                GenericReactionFactory(
                    bond_orders={
                        frozenset(pair): 9 for pair in functional_group_pairs
                    }
                )
            )

        super().__init__(
            building_block_vertices=building_block_vertices,
            edges=self._edge_prototypes,
            reaction_factory=reaction_factory,
            construction_stages=(),
            num_processes=num_processes,
            optimizer=optimizer,
            edge_groups=None,
        )

    def _normalize_metals(
        self,
        metals: typing.Union[
            BuildingBlock,
            dict[BuildingBlock, tuple[int, ...]],
            tuple[BuildingBlock, ...],
        ],
    ) -> dict[BuildingBlock, abc.Sequence[Vertex]]:
        """
        Return a map between metals and vertices.

        Parameters:

            metals:
                The metal-based building blocks.

        Returns:
            Map of :class:`.BuildingBlock` to a :class:`tuple` of
            :class:`.Vertex`

        """

        metals_dict: dict[BuildingBlock, abc.Sequence[Vertex]]

        if isinstance(metals, dict):
            metals_dict = {
                metal: tuple(self._get_metal_vertices(ids))
                for metal, ids in metals.items()
            }
        elif isinstance(metals, tuple):
            functional_group_counter = Counter(
                metal.get_num_functional_groups() for metal in metals
            )
            assert all(
                count == 1 for count in functional_group_counter.values()
            ), (
                "Cannot use a tuple when multiple metals "
                "have the same number of functional groups. "
                "Use a dictionary instead."
            )
            metals_dict = {
                metal: tuple(
                    self._get_metal_vertices(
                        self._metal_vertices_of_degree[
                            metal.get_num_functional_groups()
                        ]
                    )
                )
                for metal in metals
            }
        else:
            ids = range(len(self._metal_vertex_prototypes))
            metals_dict = {metals: tuple(self._get_metal_vertices(ids))}

        return metals_dict

    def _normalize_ligands(
        self,
        ligands: typing.Union[
            BuildingBlock,
            tuple[BuildingBlock, ...],
            dict[BuildingBlock, tuple[int, ...]],
        ],
    ) -> dict[BuildingBlock, tuple[_LigandVertex, ...]]:
        """
        Return a map ligands and vertices.

        Parameters:

            ligands:
                The organic-based building blocks.

        Returns:

            Map of :class:`.BuildingBlock` to a :class:`tuple` of
            :class:`.Vertex`

        Raises:

            :class:`AssertionError`
                If a :class:`tuple` is provided for ligands but there
                is ambiguity on ligand-vertex assignment because two
                ligands have the same number of functional groups.

        """

        if isinstance(ligands, dict):
            ligands_dict = {
                ligand: tuple(self._get_ligand_vertices(ids))
                for ligand, ids in ligands.items()
            }
        elif isinstance(ligands, tuple):
            functional_group_counter = Counter(
                ligand.get_num_functional_groups() for ligand in ligands
            )
            assert all(
                count == 1 for count in functional_group_counter.values()
            ), (
                "Cannot use a tuple when multiple ligands "
                "have the same number of functional groups. "
                "Use a dictionary instead."
            )
            ligands_dict = {
                ligand: tuple(
                    self._get_ligand_vertices(
                        self._ligand_vertices_of_degree[
                            ligand.get_num_functional_groups()
                        ]
                    )
                )
                for ligand in ligands
            }

        else:
            ids = range(len(self._ligand_vertex_prototypes))
            ligands_dict = {ligands: tuple(self._get_ligand_vertices(ids))}

        return ligands_dict

    def _get_metal_vertices(
        self,
        vertex_ids: OneOrMany[int],
    ) -> typing.Iterator[MetalVertex]:
        """
        Yield vertex prototypes.

        Parameters:

            vertex_ids:
                The ids of the vertices to yield.

        Yields:

            A vertex prototype of the topology graph.

        """

        if isinstance(vertex_ids, int):
            vertex_ids = (vertex_ids,)

        for vertex_id in vertex_ids:
            yield self._metal_vertex_prototypes[vertex_id]

    def _get_ligand_vertices(
        self,
        vertex_ids: OneOrMany[int],
    ) -> typing.Iterator[_LigandVertex]:
        """
        Yield vertex prototypes.

        Parameters:

            vertex_ids:
                The ids of the vertices to yield.

        Yields:

            A vertex prototype of the topology graph.

        """

        if isinstance(vertex_ids, int):
            vertex_ids = (vertex_ids,)

        for vertex_id in vertex_ids:
            yield self._ligand_vertex_prototypes[vertex_id]

    def _get_scale(
        self,
        building_block_vertices: dict[BuildingBlock, abc.Sequence[Vertex]],
    ) -> float:
        return 1

    def clone(self) -> MetalComplex:
        return self._clone()

    def with_building_blocks(
        self,
        building_block_map: dict[BuildingBlock, BuildingBlock],
    ) -> MetalComplex:
        return self.clone()._with_building_blocks(building_block_map)

    def __repr__(self) -> str:
        return f"metal_complex.{self.__class__.__name__}" f"()"



import pytest
import stk

from ..case_data import CaseData


@pytest.fixture
def ring_amine(get_atom_ids):
    a, b, c, d, e, f, g = get_atom_ids(7)
    return _ring_amine(
        nitrogen=stk.N(a),
        hydrogen1=stk.H(b),
        hydrogen2=stk.H(c),
        carbon1=stk.C(d),
        carbon2=stk.C(e),
        hydrogen3=stk.H(f),
        carbon3=stk.C(g),
    )


def _ring_amine(
    nitrogen,
    hydrogen1,
    hydrogen2,
    carbon1,
    carbon2,
    hydrogen3,
    carbon3,
):
    return CaseData(
        functional_group=stk.RingAmine(
            nitrogen=nitrogen,
            hydrogen1=hydrogen1,
            hydrogen2=hydrogen2,
            carbon1=carbon1,
            carbon2=carbon2,
            hydrogen3=hydrogen3,
            carbon3=carbon3,
        ),
        atoms=(
            nitrogen,
            hydrogen1,
            hydrogen2,
            carbon1,
            carbon2,
            hydrogen3,
            carbon3,
        ),
        placers=(nitrogen,),
        core_atoms=(nitrogen,),
    )



from .add import *  # noqa
from .divide_by_mean import *  # noqa
from .multiply import *  # noqa
from .null import *  # noqa
from .power import *  # noqa
from .replace_fitness import *  # noqa
from .sequence import *  # noqa
from .shift_up import *  # noqa
from .sum import *  # noqa



import rdkit.Chem.AllChem as rdkit

from stk._internal.topology_graphs.vertex import Vertex


class AxleVertex(Vertex):
    def place_building_block(self, building_block, edges):
        return building_block.with_centroid(
            position=self._position,
            atom_ids=building_block.get_placer_ids(),
        ).get_position_matrix()

    def map_functional_groups_to_edges(self, building_block, edges):
        return {}


class CycleVertex(Vertex):
    """
    Places the cycles in a :class:`NRotaxane`.

    """

    def __init__(self, id, position, flip):
        """
        Initialize a :class:`.CycleVertex` instance.

        Parameters
        ----------
        id : :class:`int`
            The id of the vertex.

        position : :class:`tuple` of :class:`float`
            The position of the vertex.

        flip : :class:`bool`
            If ``True``, the orientation of building blocks placed by
            the vertex will be flipped.

        """

        self._flip = flip
        super().__init__(id, position)

    def get_flip(self):
        """
        Return ``True`` if the vertex flips building blocks it places.

        Returns
        -------
        :class:`bool`
            ``True`` if the vertex flips building blocks it places.

        """

        return self._flip

    def clone(self):
        clone = super().clone()
        clone._flip = self._flip
        return clone

    def place_building_block(self, building_block, edges):
        rdkit_mol = building_block.to_rdkit_mol()
        macrocycle = max(rdkit.GetSymmSSSR(rdkit_mol), key=len)
        return (
            building_block.with_centroid(
                position=self._position,
                atom_ids=macrocycle,
            )
            .with_rotation_between_vectors(
                start=building_block.get_plane_normal(macrocycle),
                target=[-1 if self._flip else 1, 0, 0],
                origin=self._position,
            )
            .get_position_matrix()
        )

    def map_functional_groups_to_edges(self, building_block, edges):
        return {}

    def __str__(self):
        return (
            f"Vertex(id={self._id}, "
            f"position={self._position.tolist()}, "
            f"flip={self._flip})"
        )



import itertools as it


def test_get_bonders(generic_case_data):
    """
    Test :meth:`.GenericFunctionalGroup.get_bonders`.

    Parameters
    ----------
    generic_case_data : :class:`.GenericCaseData`
        The test case. Holds the functional group to test and the
        correct bonder atoms.

    Returns
    -------
    None : :class:`NoneType`

    """

    _test_get_bonders(
        functional_group=generic_case_data.functional_group,
        bonders=generic_case_data.bonders,
    )


def _test_get_bonders(functional_group, bonders):
    """
    Test :meth:`.GenericFunctionalGroup.get_bonders`.

    Parameters
    ----------
    functional_group : :class:`.GenericFunctionalGroup`
        The functional group to test.

    bonders : :class:`tuple` of :class:`.Atom`
        The correct bonder atoms.

    Returns
    -------
    None : :class:`NoneType`

    """

    for atom1, atom2 in it.zip_longest(
        functional_group.get_bonders(),
        bonders,
    ):
        assert atom1 is atom2



from typing import Callable

import pytest
import stk


@pytest.fixture(
    params=[
        lambda bond: {},
        lambda bond: {bond.get_atom1().get_id(): 33},
        lambda bond: {bond.get_atom2().get_id(): 122},
        lambda bond: {
            bond.get_atom1().get_id(): 4,
            bond.get_atom2().get_id(): 7,
        },
    ],
)
def get_id_map(request) -> Callable[[stk.Bond], dict[int, int]]:
    """
    Return a valid `atom_map` parameter for a bond.

    Parameters:

        bond:
            The bond, for which the `atom_map` parameter needs to be
            created.

    Returns:

        A valid `id_map` parameter for :meth:`.Bond.id_atoms`.

    """

    return request.param



import pytest
import stk

from ..case_data import CaseData


@pytest.fixture(
    scope="session",
    params=(
        lambda: CaseData(
            factory=stk.PrimaryAminoFactory(),
            molecule=stk.BuildingBlock("NCCN"),
            functional_groups=(
                stk.PrimaryAmino(
                    nitrogen=stk.N(0),
                    hydrogen1=stk.H(4),
                    hydrogen2=stk.H(5),
                    atom=stk.C(1),
                    bonders=(stk.N(0),),
                    deleters=(stk.H(4), stk.H(5)),
                ),
                stk.PrimaryAmino(
                    nitrogen=stk.N(3),
                    hydrogen1=stk.H(10),
                    hydrogen2=stk.H(11),
                    atom=stk.C(2),
                    bonders=(stk.N(3),),
                    deleters=(stk.H(10), stk.H(11)),
                ),
            ),
        ),
        lambda: CaseData(
            factory=stk.PrimaryAminoFactory(),
            molecule=stk.BuildingBlock("CCCC"),
            functional_groups=(),
        ),
        lambda: CaseData(
            factory=stk.SecondaryAminoFactory(),
            molecule=stk.BuildingBlock("CNCCNCC"),
            functional_groups=(
                stk.SecondaryAmino(
                    nitrogen=stk.N(1),
                    hydrogen=stk.H(10),
                    atom1=stk.C(0),
                    atom2=stk.C(2),
                    bonders=(stk.N(1),),
                    deleters=(stk.H(10),),
                ),
                stk.SecondaryAmino(
                    nitrogen=stk.N(4),
                    hydrogen=stk.H(15),
                    atom1=stk.C(3),
                    atom2=stk.C(5),
                    bonders=(stk.N(4),),
                    deleters=(stk.H(15),),
                ),
            ),
        ),
        lambda: CaseData(
            factory=stk.AldehydeFactory(),
            molecule=stk.BuildingBlock("O=CCC=O"),
            functional_groups=(
                stk.Aldehyde(
                    carbon=stk.C(1),
                    oxygen=stk.O(0),
                    hydrogen=stk.H(5),
                    atom=stk.C(2),
                    bonders=(stk.C(1),),
                    deleters=(stk.O(0),),
                ),
                stk.Aldehyde(
                    carbon=stk.C(3),
                    oxygen=stk.O(4),
                    hydrogen=stk.H(8),
                    atom=stk.C(2),
                    bonders=(stk.C(3),),
                    deleters=(stk.O(4),),
                ),
            ),
        ),
        lambda: CaseData(
            factory=stk.CarboxylicAcidFactory(),
            molecule=stk.BuildingBlock("O=C(O)CCCC(O)=O"),
            functional_groups=(
                stk.CarboxylicAcid(
                    carbon=stk.C(1),
                    oxygen1=stk.O(0),
                    oxygen2=stk.O(2),
                    hydrogen=stk.H(9),
                    atom=stk.C(3),
                    bonders=(stk.C(1),),
                    deleters=(stk.O(2), stk.H(9)),
                ),
                stk.CarboxylicAcid(
                    carbon=stk.C(6),
                    oxygen1=stk.O(8),
                    oxygen2=stk.O(7),
                    hydrogen=stk.H(16),
                    atom=stk.C(5),
                    bonders=(stk.C(6),),
                    deleters=(stk.O(7), stk.H(16)),
                ),
            ),
        ),
        lambda: CaseData(
            factory=stk.AmideFactory(),
            molecule=stk.BuildingBlock("O=C(N)CCC(=O)N"),
            functional_groups=(
                stk.Amide(
                    oxygen=stk.O(0),
                    carbon=stk.C(1),
                    nitrogen=stk.N(2),
                    hydrogen1=stk.H(8),
                    hydrogen2=stk.H(9),
                    atom=stk.C(3),
                    bonders=(stk.C(1),),
                    deleters=(stk.N(2), stk.H(8), stk.H(9)),
                ),
                stk.Amide(
                    carbon=stk.C(5),
                    oxygen=stk.O(6),
                    nitrogen=stk.N(7),
                    hydrogen1=stk.H(14),
                    hydrogen2=stk.H(15),
                    atom=stk.C(4),
                    bonders=(stk.C(5),),
                    deleters=(stk.N(7), stk.H(14), stk.H(15)),
                ),
            ),
        ),
        lambda: CaseData(
            factory=stk.ThioacidFactory(),
            molecule=stk.BuildingBlock("O=C(S)CC(S)=O"),
            functional_groups=(
                stk.Thioacid(
                    carbon=stk.C(1),
                    oxygen=stk.O(0),
                    sulfur=stk.S(2),
                    hydrogen=stk.H(7),
                    atom=stk.C(3),
                    bonders=(stk.C(1),),
                    deleters=(stk.S(2), stk.H(7)),
                ),
                stk.Thioacid(
                    carbon=stk.C(4),
                    oxygen=stk.O(6),
                    sulfur=stk.S(5),
                    hydrogen=stk.H(10),
                    atom=stk.C(3),
                    bonders=(stk.C(4),),
                    deleters=(stk.S(5), stk.H(10)),
                ),
            ),
        ),
        lambda: CaseData(
            factory=stk.AlcoholFactory(),
            molecule=stk.BuildingBlock("OCCCO"),
            functional_groups=(
                stk.Alcohol(
                    oxygen=stk.O(0),
                    hydrogen=stk.H(5),
                    atom=stk.C(1),
                    bonders=(stk.O(0),),
                    deleters=(stk.H(5),),
                ),
                stk.Alcohol(
                    oxygen=stk.O(4),
                    hydrogen=stk.H(12),
                    atom=stk.C(3),
                    bonders=(stk.O(4),),
                    deleters=(stk.H(12),),
                ),
            ),
        ),
        lambda: CaseData(
            factory=stk.ThiolFactory(),
            molecule=stk.BuildingBlock("SCCCS"),
            functional_groups=(
                stk.Thiol(
                    sulfur=stk.S(0),
                    hydrogen=stk.H(5),
                    atom=stk.C(1),
                    bonders=(stk.S(0),),
                    deleters=(stk.H(5),),
                ),
                stk.Thiol(
                    sulfur=stk.S(4),
                    hydrogen=stk.H(12),
                    atom=stk.C(3),
                    bonders=(stk.S(4),),
                    deleters=(stk.H(12),),
                ),
            ),
        ),
        lambda: CaseData(
            factory=stk.FluoroFactory(),
            molecule=stk.BuildingBlock("FCC(F)CCF"),
            functional_groups=(
                stk.Fluoro(
                    fluorine=stk.F(0),
                    atom=stk.C(1),
                    bonders=(stk.C(1),),
                    deleters=(stk.F(0),),
                ),
                stk.Fluoro(
                    fluorine=stk.F(3),
                    atom=stk.C(2),
                    bonders=(stk.C(2),),
                    deleters=(stk.F(3),),
                ),
                stk.Fluoro(
                    fluorine=stk.F(6),
                    atom=stk.C(5),
                    bonders=(stk.C(5),),
                    deleters=(stk.F(6),),
                ),
            ),
        ),
        lambda: CaseData(
            factory=stk.BromoFactory(),
            molecule=stk.BuildingBlock("BrCC(Br)CCBr"),
            functional_groups=(
                stk.Bromo(
                    bromine=stk.Br(0),
                    atom=stk.C(1),
                    bonders=(stk.C(1),),
                    deleters=(stk.Br(0),),
                ),
                stk.Bromo(
                    bromine=stk.Br(3),
                    atom=stk.C(2),
                    bonders=(stk.C(2),),
                    deleters=(stk.Br(3),),
                ),
                stk.Bromo(
                    bromine=stk.Br(6),
                    atom=stk.C(5),
                    bonders=(stk.C(5),),
                    deleters=(stk.Br(6),),
                ),
            ),
        ),
        lambda: CaseData(
            factory=stk.IodoFactory(),
            molecule=stk.BuildingBlock("ICC(I)CCI"),
            functional_groups=(
                stk.Iodo(
                    iodine=stk.I(0),
                    atom=stk.C(1),
                    bonders=(stk.C(1),),
                    deleters=(stk.I(0),),
                ),
                stk.Iodo(
                    iodine=stk.I(3),
                    atom=stk.C(2),
                    bonders=(stk.C(2),),
                    deleters=(stk.I(3),),
                ),
                stk.Iodo(
                    iodine=stk.I(6),
                    atom=stk.C(5),
                    bonders=(stk.C(5),),
                    deleters=(stk.I(6),),
                ),
            ),
        ),
        lambda: CaseData(
            factory=stk.TerminalAlkyneFactory(),
            molecule=stk.BuildingBlock("C#CC#CC"),
            functional_groups=(
                stk.Alkyne(
                    carbon1=stk.C(0),
                    carbon2=stk.C(1),
                    atom1=stk.H(5),
                    atom2=stk.C(2),
                    bonders=(stk.C(1),),
                    deleters=(
                        stk.C(0),
                        stk.H(5),
                    ),
                ),
            ),
        ),
        lambda: CaseData(
            factory=stk.TerminalAlkeneFactory(),
            molecule=stk.BuildingBlock("C=CC=CC"),
            functional_groups=(
                stk.Alkene(
                    carbon1=stk.C(0),
                    carbon2=stk.C(1),
                    atom1=stk.H(5),
                    atom2=stk.H(6),
                    atom3=stk.C(2),
                    atom4=stk.H(7),
                    bonders=(stk.C(1),),
                    deleters=(stk.C(0), stk.H(5), stk.H(6)),
                ),
            ),
        ),
        lambda: CaseData(
            factory=stk.BoronicAcidFactory(),
            molecule=stk.BuildingBlock("B(O)(O)CCB(O)O"),
            functional_groups=(
                stk.BoronicAcid(
                    boron=stk.B(0),
                    oxygen1=stk.O(1),
                    oxygen2=stk.O(2),
                    hydrogen1=stk.H(8),
                    hydrogen2=stk.H(9),
                    atom=stk.C(3),
                    bonders=(stk.B(0),),
                    deleters=(stk.O(1), stk.O(2), stk.H(8), stk.H(9)),
                ),
                stk.BoronicAcid(
                    boron=stk.B(5),
                    oxygen1=stk.O(6),
                    oxygen2=stk.O(7),
                    hydrogen1=stk.H(14),
                    hydrogen2=stk.H(15),
                    atom=stk.C(4),
                    bonders=(stk.B(5),),
                    deleters=(
                        stk.O(6),
                        stk.O(7),
                        stk.H(14),
                        stk.H(15),
                    ),
                ),
            ),
        ),
        lambda: CaseData(
            factory=stk.DiolFactory(),
            molecule=stk.BuildingBlock("CC(O)C(O)CC"),
            functional_groups=(
                stk.Diol(
                    atom1=stk.C(1),
                    oxygen1=stk.O(2),
                    atom2=stk.C(3),
                    oxygen2=stk.O(4),
                    hydrogen1=stk.H(11),
                    hydrogen2=stk.H(13),
                    bonders=(stk.C(1), stk.C(3)),
                    deleters=(
                        stk.O(2),
                        stk.O(4),
                        stk.H(11),
                        stk.H(13),
                    ),
                ),
            ),
        ),
        lambda: CaseData(
            factory=stk.DifluoroFactory(),
            molecule=stk.BuildingBlock("CC(F)C(F)CC"),
            functional_groups=(
                stk.Difluoro(
                    atom1=stk.C(1),
                    fluorine1=stk.F(2),
                    atom2=stk.C(3),
                    fluorine2=stk.F(4),
                    bonders=(stk.C(1), stk.C(3)),
                    deleters=(stk.F(2), stk.F(4)),
                ),
            ),
        ),
        lambda: CaseData(
            factory=stk.DibromoFactory(),
            molecule=stk.BuildingBlock("CC(Br)C(Br)CC"),
            functional_groups=(
                stk.Dibromo(
                    atom1=stk.C(1),
                    bromine1=stk.Br(2),
                    atom2=stk.C(3),
                    bromine2=stk.Br(4),
                    bonders=(stk.C(1), stk.C(3)),
                    deleters=(stk.Br(2), stk.Br(4)),
                ),
            ),
        ),
        lambda: CaseData(
            factory=stk.SmartsFunctionalGroupFactory(
                smarts="[C][N]",
                bonders=(0,),
                deleters=(1,),
            ),
            molecule=stk.BuildingBlock("NCC"),
            functional_groups=(
                stk.GenericFunctionalGroup(
                    atoms=(stk.N(0), stk.C(1)),
                    bonders=(stk.C(1),),
                    deleters=(stk.N(0),),
                ),
            ),
        ),
        lambda: CaseData(
            factory=stk.SmartsFunctionalGroupFactory(
                smarts="CBr",
                bonders=(0,),
                deleters=(1,),
                placers=(1,),
            ),
            molecule=stk.BuildingBlock("CC(Br)C(Br)CC"),
            functional_groups=(
                stk.GenericFunctionalGroup(
                    atoms=(stk.C(1), stk.Br(2)),
                    bonders=(stk.C(1),),
                    deleters=(stk.Br(2),),
                    placers=(stk.Br(2),),
                ),
                stk.GenericFunctionalGroup(
                    atoms=(stk.C(3), stk.Br(4)),
                    bonders=(stk.C(3),),
                    deleters=(stk.Br(4),),
                    placers=(stk.Br(4),),
                ),
            ),
        ),
    ),
)
def case_data(request) -> CaseData:
    """
    A :class:`.CaseData` instance.

    """

    return request.param()



import numpy as np


def test_get_cell_matrix(periodic_case):
    """
    Test :meth:`.PeriodicInfo.get_cell_matrix`.

    Parameters
    ----------
    periodic_case : :class:`.CaseData`
        The periodic case to test.

    Returns
    -------
    None : :class:`NoneType`

    """

    test = np.array(
        (
            periodic_case.vector_1,
            periodic_case.vector_2,
            periodic_case.vector_3,
        )
    )
    original = np.array(periodic_case.periodic_info.get_cell_matrix())

    assert np.all(np.allclose(test, original, atol=1e-6))



import numpy as np

from ...utilities import get_num_atom_ids, normalize_ids
from ..utilities import get_direction
from .utilities import get_plane_normal


def test_get_plane_normal(case_data, get_atom_ids):
    """
    Test :meth:`.Molecule.get_plane_normal`.

    Parameters
    ----------
    case_data : :class:`.CaseData`
        A test case. Holds the molecule to test and the correct
        positions of its atoms.

    get_atom_ids : :class:`callable`
        Takes a single parameter, `molecule`, and returns a valid
        `atom_ids` parameter for :meth:`.Molecule.get_plane_normal`.
        This allows the testing of different values of this parameter.

    Returns
    -------
    None : :class:`NoneType`

    Notes
    -----
    This test compares the result of
    :meth:`.Molecule.get_plane_normal` to the result of
    :func:`.get_plane_normal`, which is a utility function defined
    for the purposes of this test. Because
    :func:`.get_plane_normal` is tested independently, in
    :mod:`.test_get_plane_normal_helper`, if its tests pass, then
    it can be assumed, that :func:`.get_plane_normal` gives
    correct results.

    Now, assuming that :func:`.get_plane_normal` passed all of its
    tests, this test compares the results of
    :meth:`.Molecule.get_plane_normal` to the results of
    :func:`.get_plane_normal`. If the results do not match, the
    fault can be placed on :meth:`.Molecule.get_plane_normal`,
    because :func:`.get_plane_normal` has already been verified to
    be correct by its own tests.

    """

    _test_get_plane_normal(
        molecule=case_data.molecule,
        get_atom_ids=get_atom_ids,
        normal=get_plane_normal(
            position_matrix=case_data.position_matrix,
            atom_ids=tuple(
                normalize_ids(
                    molecule=case_data.molecule,
                    ids=get_atom_ids(case_data.molecule),
                )
            ),
        ),
    )


def _test_get_plane_normal(molecule, get_atom_ids, normal):
    """
    Test :meth:`.Molecule.get_plane_normal`.

    Parameters
    ----------
    molecule : :class:`.Molecule`
        The molecule to test.

    get_atom_ids : :class:`callable`
        Takes a single parameter, `molecule`, and returns a valid
        `atom_ids` parameter for :meth:`.Molecule.get_plane_normal`.
        This allows the testing of different values of this parameter.

    normal : :class:`numpy.ndarray`
        The correct plane normal.

    Returns
    -------
    None : :class:`NoneType`

    """

    num_atom_ids = get_num_atom_ids(molecule, get_atom_ids)
    if num_atom_ids == 1:
        # Any non-0 vector is valid in this case.
        assert not np.allclose(
            a=[0, 0, 0],
            b=molecule.get_plane_normal(get_atom_ids(molecule)),
            atol=1e-8,
        )
        return

    elif num_atom_ids == 2:
        # Any perpendicular vector is valid in this case.
        result = molecule.get_plane_normal(get_atom_ids(molecule))
        direction = get_direction(
            position_matrix=molecule.get_position_matrix(),
            atom_ids=(0, 1),
        )
        assert result @ direction < 1e-8
        return

    result = molecule.get_plane_normal(get_atom_ids(molecule))
    # The normal may be parallel or anti-parallel.
    assert abs(abs(result @ normal) - 1) < 1e-8



import pytest
import stk

from ..case_data import GenericCaseData


@pytest.fixture
def fluoro(get_atom_ids):
    a, b = get_atom_ids(2)
    return _fluoro(stk.F(a), stk.C(b))


def _fluoro(fluorine, atom):
    bonders = (atom,)
    deleters = (fluorine,)
    return GenericCaseData(
        functional_group=stk.Fluoro(
            fluorine=fluorine,
            atom=atom,
            bonders=bonders,
            deleters=deleters,
        ),
        atoms=(fluorine, atom),
        bonders=bonders,
        deleters=deleters,
    )



def test_to_json(case_data):
    """
    Test :meth:`.to_json`.

    Parameters
    ----------
    case_data : :class:`.CaseData`
        A test case. Holds the jsonizer and the correct json of a
        molecule.

    Returns
    -------
    None : :class:`NoneType`

    """

    _test_to_json(
        jsonizer=case_data.jsonizer,
        molecule=case_data.molecule,
        json=case_data.json,
    )


def _test_to_json(jsonizer, molecule, json):
    """
    Test :meth:`.to_json`.

    Parameters
    ----------
    jsonizer : :class:`.MoleculeJsonizer` or \
            :class:`.ConstructedMoleculeJsonizer`
        The jsonizer to test.

    molecule : :class:`.Molecule`
        The molecule to JSONize.

    json : :class:`dict`
        The correct JSON of `molecule`.

    Returns
    -------
    None : :class:`NoneType`

    """

    assert jsonizer.to_json(molecule) == json



import pytest
import stk

from ..case_data import CaseData


def get_eight_plus_twelve(graph):
    return stk.cage.EightPlusTwelve(graph.get_building_blocks())


@pytest.fixture(
    scope="session",
    params=(
        lambda: CaseData(
            mutator=stk.RandomTopologyGraph(
                replacement_funcs=(get_eight_plus_twelve,),
            ),
            record=stk.MoleculeRecord(
                topology_graph=stk.cage.FourPlusSix(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles="BrCCBr",
                            functional_groups=[stk.BromoFactory()],
                        ),
                        stk.BuildingBlock(
                            smiles="BrCC(CBr)CBr",
                            functional_groups=[stk.BromoFactory()],
                        ),
                    ),
                ),
            ),
            mutation_record=stk.MutationRecord(
                molecule_record=stk.MoleculeRecord(
                    topology_graph=stk.cage.EightPlusTwelve(
                        building_blocks=(
                            stk.BuildingBlock(
                                smiles="BrCCBr",
                                functional_groups=[stk.BromoFactory()],
                            ),
                            stk.BuildingBlock(
                                smiles="BrCC(CBr)CBr",
                                functional_groups=[stk.BromoFactory()],
                            ),
                        ),
                    ),
                ),
                mutator_name="RandomTopologyGraph",
            ),
        ),
    ),
)
def random_topology_graph(request) -> CaseData:
    return request.param()



import pytest
from pytest_lazyfixture import lazy_fixture

from .metal_topologies import *  # noqa

# All fixtures must be visible for lazy_fixture() call.
from .three_plus_four import *  # noqa
from .three_plus_three import *  # noqa
from .two_plus_five import *  # noqa
from .two_plus_four import *  # noqa
from .two_plus_three import *  # noqa


@pytest.fixture(
    params=(
        lazy_fixture("cage_six_plus_eight"),
        lazy_fixture("cage_four_plus_four"),
        lazy_fixture("cage_one_plus_one"),
        lazy_fixture("cage_two_plus_two"),
        lazy_fixture("cage_twelve_plus_thirty"),
        lazy_fixture("cage_eight_plus_sixteen"),
        lazy_fixture("cage_five_plus_ten"),
        lazy_fixture("cage_four_plus_eight"),
        lazy_fixture("cage_six_plus_twelve"),
        lazy_fixture("cage_ten_plus_twenty"),
        lazy_fixture("cage_three_plus_six"),
        lazy_fixture("cage_two_plus_four"),
        lazy_fixture("cage_eight_plus_twelve"),
        lazy_fixture("cage_four_plus_six"),
        lazy_fixture("cage_four_plus_six_2"),
        lazy_fixture("cage_six_plus_nine"),
        lazy_fixture("cage_twenty_plus_thirty"),
        lazy_fixture("cage_two_plus_three"),
        lazy_fixture("metal_cage_m2l4_lantern"),
        lazy_fixture("metal_cage_m3l3_triangle"),
        lazy_fixture("metal_cage_m3l6"),
        lazy_fixture("metal_cage_m4l4_square"),
        lazy_fixture("metal_cage_m4l4_tetrahedron"),
        lazy_fixture("metal_cage_m4l6_tetrahedron"),
        lazy_fixture("metal_cage_m4l6_tetrahedron_spacer"),
        lazy_fixture("metal_cage_m4l8"),
        lazy_fixture("metal_cage_m6l2l3_prism"),
        lazy_fixture("metal_cage_m6l12_cube"),
        lazy_fixture("metal_cage_m8l6_cube"),
        lazy_fixture("metal_cage_m12l24"),
        lazy_fixture("metal_cage_m24l48"),
    ),
)
def cage(request):
    return request.param



from collections import Counter


def check_building_block_counts(old_state, new_state, building_blocks):
    counts = Counter(building_blocks)
    old_count = old_state.get_building_block_counts()
    new_count = new_state.get_building_block_counts()

    for building_block in old_count:
        expected_count = old_count[building_block] - new_count[building_block]
        assert counts.get(building_block, 0) == expected_count

    new_only = new_count.keys() - old_count.keys()
    added_only = counts.keys() - old_count.keys()
    assert len(new_only) == len(added_only)

    for building_block in new_only:
        assert new_count[building_block] == counts[building_block]



def test_get_charge(case_data):
    """
    Test :meth:`.Atom.get_charge`.

    Parameters
    ----------
    case_data : :class:`.CaseData`
        The test case. Holds the atom to test and its correct charge.

    Returns
    -------
    None : :class:`NoneType`

    """

    assert case_data.atom.get_charge() == case_data.charge



import typing

T = typing.TypeVar("T")


def get_rank_fitness(population: dict[T, float]) -> dict[T, float]:
    sorted_population = sorted(
        zip(population.values(), population.keys()),
        reverse=True,
    )
    return {
        record: 1 / rank
        for rank, (_, record) in enumerate(sorted_population, 1)
    }



from typing import Any

import pytest
import stk
from pytest_lazyfixture import lazy_fixture

from ..case_data import CaseData


def get_topology_graph(num_repeating_units: int) -> stk.TopologyGraph:
    return stk.polymer.Linear(
        building_blocks=[
            stk.BuildingBlock("BrCCBr", stk.BromoFactory()),
        ],
        repeating_unit="A",
        num_repeating_units=num_repeating_units,
    )


@pytest.fixture(scope="session")
def worst_population_1() -> dict[stk.MoleculeRecord[Any], float]:
    return {
        stk.MoleculeRecord(
            topology_graph=get_topology_graph(2),
        ): 11,
        stk.MoleculeRecord(
            topology_graph=get_topology_graph(3),
        ): 10,
        stk.MoleculeRecord(
            topology_graph=get_topology_graph(4),
        ): 9,
        stk.MoleculeRecord(
            topology_graph=get_topology_graph(5),
        ): 2,
        stk.MoleculeRecord(
            topology_graph=get_topology_graph(6),
        ): 1,
    }


@pytest.fixture(scope="session")
def worst_population_2() -> dict[stk.MoleculeRecord[Any], float]:
    return {
        stk.MoleculeRecord(
            topology_graph=get_topology_graph(6),
        ): 100,
        stk.MoleculeRecord(
            topology_graph=get_topology_graph(6),
        ): 100,
        stk.MoleculeRecord(
            topology_graph=get_topology_graph(8),
        ): 1,
    }


@pytest.fixture(
    scope="session",
    params=(
        lambda population: CaseData.new(
            selector=stk.Worst(),
            population=population,
            selected=([4], [3], [2], [1], [0]),
        ),
        lambda population: CaseData.new(
            selector=stk.Worst(num_batches=2),
            population=population,
            selected=([4], [3]),
        ),
        lambda population: CaseData.new(
            selector=stk.Worst(
                batch_size=2,
            ),
            population=population,
            selected=(
                [3, 4],
                [2, 4],
                [1, 4],
                [2, 3],
                [0, 4],
                [1, 3],
                [0, 3],
                [1, 2],
                [0, 2],
                [0, 1],
            ),
        ),
        lambda population: CaseData.new(
            selector=stk.Worst(
                num_batches=3,
                batch_size=2,
            ),
            population=population,
            selected=([3, 4], [2, 4], [1, 4]),
        ),
        lambda population: CaseData.new(
            selector=stk.Worst(
                batch_size=2,
                duplicate_molecules=False,
            ),
            population=population,
            selected=([3, 4], [1, 2]),
        ),
    ),
)
def worst_population_1_case_data(
    request: pytest.FixtureRequest,
    worst_population_1: dict[stk.MoleculeRecord[Any], float],
) -> CaseData:
    return request.param(worst_population_1)


@pytest.fixture(
    scope="session",
    params=(
        lambda population: CaseData.new(
            selector=stk.Worst(
                duplicate_molecules=False,
            ),
            population=population,
            selected=([2], [0]),
        ),
        lambda population: CaseData.new(
            selector=stk.Worst(
                duplicate_batches=False,
            ),
            population=population,
            selected=([2], [0]),
        ),
        lambda population: CaseData.new(
            selector=stk.Worst(
                batch_size=2,
                duplicate_batches=False,
            ),
            population=population,
            selected=([0, 2], [0, 1]),
        ),
    ),
)
def worst_population_2_case_data(
    request: pytest.FixtureRequest,
    worst_population_2: dict[stk.MoleculeRecord[Any], float],
) -> CaseData:
    return request.param(worst_population_2)


@pytest.fixture(
    scope="session",
    params=(
        lazy_fixture("worst_population_1_case_data"),
        lazy_fixture("worst_population_2_case_data"),
    ),
)
def worst(request: pytest.FixtureRequest) -> CaseData:
    return request.param



"""
Turbomole Writer
================

"""

from __future__ import annotations

import typing

from stk._internal.molecule import Molecule
from stk._internal.periodic_info import PeriodicInfo
from stk._internal.utilities.utilities import OneOrMany


class TurbomoleWriter:
    """
    A writer class for ``Turbomole`` files.

    Examples:

        *Writing to a File with a Unit Cell*

        This writer can write to a file with the unit
        cell included for periodic molecules. Note that this always
        assumes P1 space group.

        .. testcode:: writing-to-a-file-with-a-unit-cell

            import stk

            bb1 = stk.BuildingBlock('BrCCBr', [stk.BromoFactory()])
            bb2 = stk.BuildingBlock(
                smiles='BrCC(CBr)CBr',
                functional_groups=[stk.BromoFactory()],
            )
            topology_graph = stk.cof.PeriodicHoneycomb(
                building_blocks=(bb1, bb2),
                lattice_size=(3, 3, 1),
            )
            construction_result = topology_graph.construct()
            cof = (
                stk.ConstructedMolecule.init_from_construction_result(
                    construction_result=construction_result,
                )
            )
            writer = stk.TurbomoleWriter()
            writer.write(
                molecule=cof,
                path='cof.coord',
                periodic_info=construction_result.get_periodic_info(),
            )

        .. testcode:: writing-to-a-file-with-a-unit-cell
            :hide:

            import os

            assert os.path.exists('cof.coord')

        .. testcleanup:: writing-to-a-file-with-a-unit-cell

            os.remove('cof.coord')

    """

    def _write_content(
        self,
        molecule: Molecule,
        atom_ids: typing.Optional[OneOrMany[int]],
        periodic_info: typing.Optional[PeriodicInfo] = None,
    ) -> list[str]:
        if atom_ids is None:
            atom_ids = range(molecule.get_num_atoms())
        elif isinstance(atom_ids, int):
            atom_ids = (atom_ids,)

        content = []
        if periodic_info is not None:
            # Input unit cell information.
            a = periodic_info.get_a()
            b = periodic_info.get_b()
            c = periodic_info.get_c()
            alpha = periodic_info.get_alpha()
            beta = periodic_info.get_beta()
            gamma = periodic_info.get_gamma()
            content.append(
                "$periodic 3\n"
                "$cell angs\n"
                f" {a:>8.3f} {b:>8.3f} {c:>8.3f} "
                f"{alpha:>6.2f} {beta:>6.2f} {gamma:>6.2f}\n"
            )

        coords = molecule.get_position_matrix()
        content.append("$coord angs\n")
        for atom_id in atom_ids:
            (atom,) = molecule.get_atoms(atom_ids=atom_id)
            element = atom.__class__.__name__
            x, y, z = (i for i in coords[atom_id])
            content.append(
                f" {round(x, 4)} {round(y, 4)} {round(z, 4)} " f"{element}\n"
            )

        content.append("$end\n")

        return content

    def to_string(
        self,
        molecule: Molecule,
        atom_ids: typing.Optional[OneOrMany[int]] = None,
        periodic_info: typing.Optional[PeriodicInfo] = None,
    ) -> str:
        """
        Get a ``Turbomole`` file format string of `molecule`.

        Parameters:

            molecule:
                Molecule to write to ``Turbomole`` format.

            atom_ids:
                The atom ids of atoms to write. Can be a single
                :class:`int`, if a single atom is to be used, or
                ``None``, if all atoms are to be used.

            periodic_info:
                Information about the periodic cell.

        Returns:

            The content of a `.coord` file.

        """

        content = self._write_content(
            molecule=molecule,
            atom_ids=atom_ids,
            periodic_info=periodic_info,
        )

        return "".join(content)

    def write(
        self,
        molecule: Molecule,
        path: str,
        atom_ids: typing.Optional[OneOrMany[int]] = None,
        periodic_info: typing.Optional[PeriodicInfo] = None,
    ) -> None:
        """
        Write `molecule` to ``Turbomole`` file format.

        Parameters:

            molecule:
                Molecule to write to ``Turbomole`` format.

            path:
                The full path to the file being written.

            atom_ids:
                The atom ids of atoms to write. Can be a single
                :class:`int`, if a single atom is to be used, or
                ``None``, if all atoms are to be used.

            periodic_info:
                Information about the periodic cell.

        Returns:

            A file is written.

        """

        content = self._write_content(
            molecule=molecule,
            atom_ids=atom_ids,
            periodic_info=periodic_info,
        )

        with open(path, "w") as f:
            f.write("".join(content))



def test_get_key_name(case_data):
    """
    Test :meth:`.get_key_name`.

    Parameters
    ----------
    case_data : :class:`.CaseData`
        A test case. Holds the key maker to test and the correct name.

    Returns
    -------
    None : :class:`NoneType`

    """

    assert case_data.key_maker.get_key_name() == case_data.key_name



from functools import partial

import numpy as np
import pytest
import stk
from pytest_lazyfixture import lazy_fixture
from scipy.spatial.distance import euclidean

from ....case_data import CaseData


@pytest.fixture(
    params=(
        lazy_fixture("nonlinear_3"),
        lazy_fixture("nonlinear_4"),
    ),
)
def nonlinear(request):
    return request.param


@pytest.fixture
def nonlinear_3(position, aligner_3, building_block_3):
    return _nonlinear(position, aligner_3, building_block_3)


@pytest.fixture(params=(0, 1, 2))
def aligner_3(request):
    """
    The `aligner_edge` parameter for vertices with 3 edges.

    """

    return request.param


@pytest.fixture(
    params=(
        lambda: stk.BuildingBlock(
            smiles="BrCC(Br)CNCBr",
            functional_groups=[stk.BromoFactory()],
        ),
    ),
)
def building_block_3(request) -> stk.BuildingBlock:
    """
    A :class:`.BuildingBlock` with 3 functional groups.

    """

    return request.param()


@pytest.fixture
def nonlinear_4(position, aligner_4, building_block_4):
    return _nonlinear(position, aligner_4, building_block_4)


@pytest.fixture(params=(0, 1, 2, 3))
def aligner_4(request):
    """
    The `aligner_edge` parameter for vertices with 4 edges.

    """

    return request.param


@pytest.fixture(
    params=(
        lambda: stk.BuildingBlock(
            smiles="BrC(CCCCCC)C(Br)CC(Br)CNCBr",
            functional_groups=[stk.BromoFactory()],
        ),
    ),
)
def building_block_4(request) -> stk.BuildingBlock:
    """
    A :class:`.BuildingBlock` with 4 functional groups.

    """

    return request.param()


@pytest.fixture(
    params=(
        [0, 0, 0],
        [1, 2, -20],
    ),
)
def position(request):
    """
    The `position` of a vertex.

    """

    return np.array(request.param, dtype=np.float64)


def _nonlinear(position, aligner_edge, building_block):
    """
    Return a test case for a nonlinear COF vertex.

    Parameters
    ----------
    position : :class:`numpy.ndarray`
        The position of the vertex.

    aligner_edge : :class:`int`
        The aligner edge of the vertex.

    building_block : :class:`.BuildingBlock`
        The building block placed on the vertex.

    Returns
    -------
    :class:`.CaseData`
        The test case.

    """

    building_block = order_functional_groups(building_block)
    n = building_block.get_num_functional_groups()
    points = tuple(get_points(position, n))
    alignment_tests = {
        partial(get_fg_point, points, 0): points[aligner_edge],
        get_normal: np.array([0, 0, 1]),
    }

    vertex = stk.cof.NonLinearVertex(
        id=0,
        position=position,
        aligner_edge=aligner_edge,
        cell=[0, 0, 0],
    )

    return CaseData(
        vertex=vertex,
        edges=tuple(get_nonlinear_edges(n, vertex)),
        building_block=building_block,
        position=position,
        alignment_tests=alignment_tests,
        functional_group_edges={
            fg_id: (fg_id + aligner_edge) % n for fg_id in range(n)
        },
    )


def get_points(center, num_points):
    """
    Yield equally spaced points on a circle of radius 10.

    Parameters
    ----------
    center : :class:`numpy.ndarray`
        The center of circle.

    num_points : :class:`int`
        The number of points to make.

    Yields
    ------
    :class:`numpy.ndarray`
        A point on a circle.

    """

    # Take slice to account for case where rounding errors cause
    # extra theta.
    thetas = np.arange(0, 2 * np.pi, 2 * np.pi / num_points)[:num_points]
    for theta in thetas:
        yield center + 10 * np.array([np.sin(theta), np.cos(theta), 0.0])


def get_fg_point(points, fg_id, building_block):
    """
    Get the point in `points` closest to a functional group.

    Parameters
    ----------
    points : :class:`iterable` of :class:`numpy.ndarray`
        The points from which the closest one is picked.

    fg_id : :class:`int`
        The id of a functional group of `building_block` whose
        distance to each point in `points` is evaluated.

    building_block : :class:`.BuildingBlock`
        The building block which owns the functional group.

    Returns
    -------
    :class:`numpy.ndarray`
        The point in `points` closest to functional group `fg_id`.

    """

    fg = next(building_block.get_functional_groups(fg_id))
    fg_position = building_block.get_centroid(
        atom_ids=fg.get_placer_ids(),
    )
    return min(points, key=partial(euclidean, fg_position))


def get_normal(building_block):
    """
    Get a normal to a plane crossing the building block.

    The normal is defined by the plane of best fit to the placer atoms
    of the building block. It is defined such that the it always forms
    an acute angle with the vector running from the centroid of the
    placer atoms to the centroid of the building block.

    Parameters
    ----------
    building_block : :.BuildingBlock`
        The building block whose normal should be calculated.

    Returns
    -------
    :class:`numpy.ndarray`
        The normal of `building_block`.

    """

    placer_centroid = building_block.get_centroid(
        atom_ids=building_block.get_placer_ids(),
    )
    normal = stk.get_acute_vector(
        reference=building_block.get_centroid() - placer_centroid,
        vector=building_block.get_plane_normal(
            atom_ids=building_block.get_placer_ids(),
        ),
    )
    if np.allclose(normal, [0, 0, 1], atol=1e-13):
        return np.array([0, 0, 1])
    return normal


def get_nonlinear_edges(num_edges, vertex):
    """
    Yield edges placed in a circle around `vertex`.

    Parameters
    ----------
    num_edges : :class:`int`
        The number of edges to yield.

    vertex : :class:`.Vertex`
        The vertex which needs edges.

    Yields
    ------
    :class:`.Edge`
        An edge connected to `vertex`.

    """

    for id_, point in enumerate(
        get_points(center=vertex.get_position(), num_points=num_edges)
    ):
        yield stk.cof.Edge(id_, id_, vertex, stk.Vertex(id_ + 1, point))


def order_functional_groups(building_block):
    """
    Get a building block with ordered functional groups.

    Parameters
    ----------
    building_block : :class:`.BuildingBlock`
        The building block whose functional groups should be
        ordered.

    Returns
    -------
    :class:`.BuildingBlock`
        A building block with ordered functional groups.

    """

    return building_block.with_functional_groups(
        functional_groups=sorted(
            building_block.get_functional_groups(),
            key=functional_group_angle(building_block),
        ),
    )


def functional_group_angle(building_block):
    fg0 = next(building_block.get_functional_groups(0))
    fg0_position = building_block.get_centroid(
        atom_ids=fg0.get_placer_ids(),
    )
    placer_centroid = building_block.get_centroid(
        atom_ids=building_block.get_placer_ids(),
    )
    fg0_direction = fg0_position - placer_centroid
    centroid = building_block.get_centroid(
        atom_ids=building_block.get_placer_ids(),
    )
    normal = stk.get_acute_vector(
        reference=building_block.get_centroid() - placer_centroid,
        vector=building_block.get_plane_normal(
            atom_ids=building_block.get_placer_ids(),
        ),
    )
    axis = -np.cross(normal, fg0_direction)

    def inner(functional_group):
        position = building_block.get_centroid(
            atom_ids=functional_group.get_placer_ids(),
        )
        fg_direction = position - centroid
        theta = stk.vector_angle(fg0_direction, fg_direction)

        projection = fg_direction @ axis
        if theta > 0 and projection < 0:
            return 2 * np.pi - theta
        return theta

    return inner



class CaseData:
    """
    A test case.

    Attributes
    ----------
    dejsonizer : :class:`.MoleculeDejsonizer`
        The dejsonizer to test.

    json : :class:`dict`
        The JSON to test.

    molecule : :class:`.Molecule`
        The correct dejsonized molecule.

    """

    def __init__(self, dejsonizer, json, molecule):
        """
        Initialize a :class:`.CaseData` instance.

        Parameters
        ----------
        dejsonizer : :class:`.MoleculeDejsonizer`
            The dejsonizer to test.

        json : :class:`dict`
            The JSON to test.

        molecule : :class:`.Molecule`
            The correct dejsonized molecule.

        """

        self.dejsonizer = dejsonizer
        self.json = json
        self.molecule = molecule



"""
Construction Result
===================

"""


class ConstructionResult:
    """
    The result of :meth:`.TopologyGraph.construct`.

    """

    __slots__ = [
        "_atoms",
        "_bonds",
        "_atom_infos",
        "_bond_infos",
        "_position_matrix",
        "_num_building_blocks",
    ]

    def __init__(self, construction_state):
        """
        Initialize a :class:`.ConstructionResult`.

        Parameters
        ----------
        construction_state : :class:`.ConstructionState`
            The state from which the result is initialized.

        """

        self._position_matrix = construction_state.get_position_matrix()
        self._position_matrix.setflags(write=False)
        self._atoms = tuple(construction_state.get_atoms())
        self._bonds = tuple(construction_state.get_bonds())
        self._atom_infos = tuple(construction_state.get_atom_infos())
        self._bond_infos = tuple(construction_state.get_bond_infos())
        self._num_building_blocks = {
            building_block: construction_state.get_num_building_block(
                building_block=building_block,
            )
            for building_block in construction_state.get_building_blocks()
        }

    def get_position_matrix(self):
        """
        Get the position matrix of the constructed molecule.

        Returns
        -------
        :class:`numpy.ndarray`
            The position matrix of the constructed molecule.

        """

        return self._position_matrix

    def get_atoms(self):
        """
        Get the atoms of the constructed molecule.

        Returns
        -------
        :class:`tuple` of :class:`.Atom`
            The atoms of the constructed molecule.

        """

        return self._atoms

    def get_bonds(self):
        """
        Get the bonds of the constructed molecule.

        Returns
        -------
        :class:`tuple` of :class:`.Bond`
            The bonds of the constructed molecule.

        """

        return self._bonds

    def get_atom_infos(self):
        """
        Get the atom infos of the constructed molecule.

        Returns
        -------
        :class:`tuple` of :class:`.AtomInfo`
            The atom infos of the constructed molecule.

        """

        return self._atom_infos

    def get_bond_infos(self):
        """
        Get the bond infos of the constructed molecule.

        Returns
        -------
        :class:`tuple` of :class:`.BondInfo`
            The bond infos of the constructed molecule.

        """

        return self._bond_infos

    def get_num_building_block(self, building_block):
        """
        Get the number of times `building_block` is present.

        Parameters
        ----------
        building_block : :class:`.BuildingBlock`
            The building block whose frequency in the constructed
            molecule is desired.

        Returns
        -------
        :class:`int`
            The number of times `building_block` was used in the
            construction of the constructed molecule.

        """

        return self._num_building_blocks[building_block]

    def get_building_blocks(self):
        """
        Yield the building blocks.

        Building blocks are yielded in an order based on their
        position in the constructed molecule. For two topologically
        equivalent constructed molecules, but with different building
        blocks, equivalently positioned building blocks will be
        yielded at the same time.

        Yields
        ------
        :class:`.BuildingBlock`
            A building block of the topology graph.

        """

        yield from self._num_building_blocks



"""
Alkene
======

"""

from __future__ import annotations

import typing

from stk._internal.atom import Atom
from stk._internal.elements import C

from .generic_functional_group import GenericFunctionalGroup


class Alkene(GenericFunctionalGroup):
    """
    Represents an alkene functional group.

    The structure of the functional group is given by the pseudo-SMILES
    ``[carbon1]([atom1])([atom2])=[carbon2]([atom3])[atom4]``.

    """

    def __init__(
        self,
        carbon1: C,
        atom1: Atom,
        atom2: Atom,
        carbon2: C,
        atom3: Atom,
        atom4: Atom,
        bonders: tuple[Atom, ...],
        deleters: tuple[Atom, ...],
        placers: typing.Optional[tuple[Atom, ...]] = None,
    ) -> None:
        """
        Initialize a :class:`.Alkene` instance.

        Parameters:

            carbon1:
                The ``[carbon1]`` atom.

            atom1:
                The ``[atom1]`` atom.

            atom2:
                The ``[atom2]`` atom.

            carbon2:
                The ``[carbon2]`` atom.

            atom3:
                The ``[atom3]`` atom.

            atom4:
                The ``[atom4]`` atom.

            bonders:
                The bonder atoms.

            deleters:
                The deleter atoms.

            placers:
                The placer atoms. If ``None`` the `bonders` will be
                used.

        """

        self._carbon1 = carbon1
        self._atom1 = atom1
        self._atom2 = atom2
        self._carbon2 = carbon2
        self._atom3 = atom3
        self._atom4 = atom4
        atoms = (carbon1, atom1, atom2, carbon2, atom3, atom4)
        super().__init__(
            atoms=atoms,
            bonders=bonders,
            deleters=deleters,
            placers=bonders if placers is None else placers,
        )

    def get_carbon1(self) -> C:
        """
        Get the ``[carbon1]`` atom.

        Returns:

            The ``[carbon1]`` atom.

        """

        return self._carbon1

    def get_atom1(self) -> Atom:
        """
        Get the ``[atom1]`` atom.

        Returns:

            The ``[atom1]`` atom.

        """

        return self._atom1

    def get_atom2(self) -> Atom:
        """
        Get the ``[atom2]`` atom.

        Returns:

            The ``[atom2]`` atom.

        """

        return self._atom2

    def get_carbon2(self) -> C:
        """
        Get the ``[carbon2]`` atom.

        Returns:

            The ``[carbon2]`` atom.

        """

        return self._carbon2

    def get_atom3(self) -> Atom:
        """
        Get the ``[atom3]`` atom.

        Returns
        -------
        :class:`.Atom`
            The ``[atom3]`` atom.

        """

        return self._atom3

    def get_atom4(self) -> Atom:
        """
        Get the ``[atom4]`` atom.

        Returns:

            The ``[atom4]`` atom.

        """

        return self._atom4

    def clone(self) -> Alkene:
        clone = super()._clone()
        clone._carbon1 = self._carbon1
        clone._atom1 = self._atom1
        clone._atom2 = self._atom2
        clone._carbon2 = self._carbon2
        clone._atom3 = self._atom3
        clone._atom4 = self._atom4
        return clone

    def _with_ids(self, id_map: dict[int, int]) -> Alkene:
        super()._with_ids(id_map)

        if (carbon1_id := self._carbon1.get_id()) in id_map:
            self._carbon1 = self._carbon1.with_id(id_map[carbon1_id])

        if (atom1_id := self._atom1.get_id()) in id_map:
            self._atom1 = self._atom1.with_id(id_map[atom1_id])

        if (atom2_id := self._atom2.get_id()) in id_map:
            self._atom2 = self._atom2.with_id(id_map[atom2_id])

        if (carbon2_id := self._carbon2.get_id()) in id_map:
            self._carbon2 = self._carbon2.with_id(id_map[carbon2_id])

        if (atom3_id := self._atom3.get_id()) in id_map:
            self._atom3 = self._atom3.with_id(id_map[atom3_id])

        if (atom4_id := self._atom4.get_id()) in id_map:
            self._atom4 = self._atom4.with_id(id_map[atom4_id])

        return self

    def with_ids(self, id_map: dict[int, int]) -> Alkene:
        return self.clone()._with_ids(id_map)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"{self._carbon1}, {self._atom1}, {self._atom2}, "
            f"{self._carbon2}, {self._atom3}, {self._atom4}, "
            f"bonders={self._bonders}, deleters={self._deleters})"
        )



def test_get_fitness_value(case_data):
    """
    Test :meth:`.FitnessCalculator.get_fitness_value`.

    Parameters
    ----------
    case_data : :class:`.CaseData`
        A test case. Holds the fitness calculator to test and the
        correct fitness value.

    Returns
    -------
    None : :class:`NoneType`

    """

    _test_get_fitness_value(
        fitness_calculator=case_data.fitness_calculator,
        molecule=case_data.molecule,
        fitness_value=case_data.fitness_value,
    )


def _test_get_fitness_value(
    fitness_calculator,
    molecule,
    fitness_value,
):
    """
    Test :meth:`.FitnessCalculator.get_fitness_value`.

    Parameters
    ----------
    fitness_calculator : :class:`.FitnessCalculator`
        The fitness calculator to test.

    molecule : :class:`.Molecule`
        The molecule whose fitness is calculated.

    fitness_value : :class:`object`
        The correct fitness value.

    Returns
    -------
    None : :class:`NoneType`

    """

    assert fitness_calculator.get_fitness_value(molecule) == fitness_value



"""
Molecule State
==============

"""

import numpy as np

from .deletions_summary import _DeletionsSummary
from .placements_summary import _PlacementsSummary
from .reactions_summary import _ReactionsSummary


class MoleculeState:
    """
    Represents the state of a molecule under construction.

    """

    __slots__ = [
        "_position_matrix",
        "_atoms",
        "_atom_infos",
        "_bonds",
        "_bond_infos",
        "_edge_functional_groups",
        "_num_placements",
    ]

    def __init__(self):
        """
        Initialize a :class:`._MoleculeState` instance.

        """

        self._position_matrix = np.empty((0, 3), dtype=np.float64)
        self._atoms = []
        self._atom_infos = []
        self._bonds = []
        self._bond_infos = []
        self._edge_functional_groups = {}
        self._num_placements = 0

    def clone(self):
        """
        Return a clone.

        Returns
        -------
        :class:`._MoleculeState`
            The clone. Has the same type as the original instance.

        """

        clone = self.__class__.__new__(self.__class__)
        clone._position_matrix = np.array(self._position_matrix)
        clone._atoms = list(self._atoms)
        clone._atom_infos = list(self._atom_infos)
        clone._bonds = list(self._bonds)
        clone._bond_infos = list(self._bond_infos)
        clone._num_placements = self._num_placements

        edge_functional_groups = self._edge_functional_groups.items()
        clone._edge_functional_groups = {
            edge_id: list(functional_groups)
            for edge_id, functional_groups in edge_functional_groups
        }
        return clone

    def _with_placement_results(
        self,
        vertices,
        edges,
        building_blocks,
        results,
    ):
        """
        Modify the instance.

        """

        summary = _PlacementsSummary(
            building_blocks=building_blocks,
            placement_results=results,
            num_atoms=len(self._atoms),
            num_previous_placements=self._num_placements,
        )
        self._num_placements += len(vertices)
        self._atoms.extend(summary.get_atoms())
        self._atom_infos.extend(summary.get_atom_infos())
        self._bonds.extend(summary.get_bonds())
        self._bond_infos.extend(summary.get_bond_infos())
        self._position_matrix = np.concatenate(
            [
                self._position_matrix,
                summary.get_position_matrix(),
            ]
        )
        for (
            edge_id,
            functional_groups,
        ) in summary.get_edge_functional_groups():
            self._edge_functional_groups[
                edge_id
            ] = self._edge_functional_groups.get(edge_id, [])
            self._edge_functional_groups[edge_id].extend(functional_groups)
        return self

    def with_placement_results(
        self,
        vertices,
        edges,
        building_blocks,
        results,
    ):
        """
        Return a clone holding the placement results.

        Parameters
        ----------
        vertices : :class:`tuple` of :class:`.Vertex`
            The vertices used for placement.

        edges : :class:`tuple`
            For each vertex in `vertices`, a :class:`tuple` of
            :class:`.Edge` instances connected to it.

        building_blocks : :class:`tuple` of :class:`.BuildingBlock`
            For each vertex in `vertices`, the building block placed
            on it.

        results : :class:`tuple` of :class:`._PlacementResult`
            For every vertex in `vertices`, the result of the
            placement.

        Returns
        -------
        :class:`._MoleculeState`
            The clone holding the placement results. Has the same
            type as the original instance.

        """

        return self.clone()._with_placement_results(
            vertices=vertices,
            edges=edges,
            building_blocks=building_blocks,
            results=results,
        )

    def get_position_matrix(self):
        """
        Get the position matrix of the molecule.

        Returns
        -------
        :class:`numpy.ndarray`
            The position matrix.

        """

        return np.array(self._position_matrix)

    def get_atoms(self):
        """
        Yield the atoms of the molecule.

        Yields
        ------
        :class:`.Atom`
            An atom of the molecule.

        """

        yield from self._atoms

    def get_bonds(self):
        """
        Yield the bonds of the molecule.

        Yields
        ------
        :class:`.Bond`
            A bond of the molecule.

        """

        yield from self._bonds

    def get_atom_infos(self):
        """
        Yield the atom infos of the molecule.

        Yields
        ------
        :class:`.AtomInfo`
            An atom info of the molecule.

        """

        yield from self._atom_infos

    def get_bond_infos(self):
        """
        Yield the bond infos of the molecule.

        Yields
        ------
        :class:`.BondInfo`
            The bond info of the molecule.

        """

        yield from self._bond_infos

    def get_edge_group_functional_groups(self, edge_group):
        """
        Yield the functional groups associated with `edge_group`.

        Parameters
        ----------
        edge_group : :class:`.EdgeGroup`
            The edge group, whose functional groups are desired.

        Yields
        ------
        :class:`.FunctionalGroup`
            A functional group which belongs to `edge_group`.

        """

        for edge_id in edge_group.get_edge_ids():
            yield from self._edge_functional_groups[edge_id]

    def with_position_matrix(self, position_matrix):
        """
        Return a clone holding the `position_matrix`.

        Parameters
        ----------
        position_matrix : :class:`numpy.ndarray`
            The position matrix of the clone. The shape of the matrix
            is ``(n, 3)``.

        Returns
        -------
        :class:`._MoleculeState`
            The clone holding the new position matrix. Has the same
            type as the original instance.

        """

        return self.clone()._with_position_matrix(position_matrix)

    def _with_position_matrix(self, position_matrix):
        """
        Modify the instance.

        """

        self._position_matrix = np.array(position_matrix)
        return self

    def with_reaction_results(self, reactions, results):
        """
        Return a clone holding the reaction results.

        Parameters
        ----------
        reactions : :class:`tuple` of :class:`.Reaction`
            The reactions.

        results : :class:`.ReactionResult`
            For each reaction in `reactions`, its result.

        Returns
        -------
        :class:`._MoleculeState`
            The clone holding the reaction results. Has the same type
            as the original instance.

        """

        return self.clone()._with_reaction_results(reactions, results)

    def _with_reaction_results(self, reactions, results):
        """
        Modify the instance.

        """

        reactions_summary = _ReactionsSummary(
            num_atoms=len(self._atoms),
            reaction_results=results,
        )
        self._with_reactions_summary(reactions_summary)
        self._with_deletions_summary(
            _DeletionsSummary(
                atoms=self._atoms,
                atom_infos=self._atom_infos,
                bonds=self._bonds,
                bond_infos=self._bond_infos,
                position_matrix=self._position_matrix,
                deleted_atom_ids=reactions_summary.get_deleted_atom_ids(),
                deleted_bond_ids=reactions_summary.get_deleted_bond_ids(),
            )
        )
        return self

    def _with_reactions_summary(self, summary):
        """
        Add the results held in `summary`.

        Parameters
        ----------
        summary : :class:`._ReactionsSummary`
            A summary of the reaction results.

        Returns
        -------
        None : :class:`NoneType`

        """

        self._atoms.extend(summary.get_atoms())
        self._atom_infos.extend(summary.get_atom_infos())
        self._bonds.extend(summary.get_bonds())
        self._bond_infos.extend(summary.get_bond_infos())

        positions = tuple(summary.get_positions())
        if positions:
            self._position_matrix = np.vstack(
                [
                    self._position_matrix,
                    positions,
                ]
            )

    def _with_deletions_summary(self, summary):
        """
        Add the results held in `summary`.

        Parameters
        ----------
        summary : :class:`._DeletionsSummary`
            A summary of the atom deletion results.

        Returns
        -------
        None : :class:`NoneType`

        """

        self._atoms = list(summary.get_atoms())
        self._atom_infos = list(summary.get_atom_infos())
        self._bonds = list(summary.get_bonds())
        self._bond_infos = list(summary.get_bond_infos())
        self._position_matrix = np.array(list(summary.get_positions()))
        return self



from collections.abc import Iterable

import numpy as np

from stk._internal.atom_info import AtomInfo
from stk._internal.bond_info import BondInfo
from stk._internal.constructed_molecule import ConstructedMolecule
from stk._internal.key_makers.inchi_key import InchiKey
from stk._internal.key_makers.molecule import MoleculeKeyMaker
from stk._internal.molecule import Molecule

from .molecule import MoleculeDejsonizer, MoleculeJsonizer
from .utilities import to_atom, to_atom_info, to_bond, to_bond_info


class ConstructedMoleculeJsonizer:
    """
    Abstract base class for creating JSONs of constructed molecules.

    See Also:
        :class:`.MoleculeJsonizer`

    Notes:
        You might notice that the public methods of this abstract base
        class are implemented. These are just default implementations,
        which can be safely ignored or overridden, when implementing
        subclasses. However, the default implementation can be used
        directly, if it suits your needs.

    Examples:
        *Converting a Constructed Molecule to JSON*

        You want get a JSON representation of a
        :class:`.ConstructedMolecule`

        .. testcode:: converting-a-molecule-to-json

            import stk

            # Make the molecule you want jsonize.
            polymer = stk.ConstructedMolecule(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(
                        stk.BuildingBlock('BrCCBr', [stk.BromoFactory()]),
                    ),
                    repeating_unit='A',
                    num_repeating_units=3,
                )
            )

            # Make a JSONizer.
            jsonizer = stk.ConstructedMoleculeJsonizer()
            # Get the JSON.
            json = jsonizer.to_json(polymer)

        *Adding Additional Molecular Keys*

        Apart from atoms, bonds and the position matrix, the JSON
        representation holds additional fields, one for each
        :class:`.MoleculeKeyMaker` provided to the initializer

        .. testcode:: adding-additional-molecular-keys

            import stk

            # Make the molecule you want jsonize.
            polymer = stk.ConstructedMolecule(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(
                        stk.BuildingBlock('BrCCBr', [stk.BromoFactory()]),
                    ),
                    repeating_unit='A',
                    num_repeating_units=3,
                )
            )

            # Make a JSONizer.
            jsonizer = stk.ConstructedMoleculeJsonizer()
            # Get the JSON.
            json = jsonizer.to_json(polymer)

        In this case, ``json`` will look something like

        .. code-block:: python

            {
                # A tuple of JSON atom representations.
                'atoms': (...),

                # A tuple of JSON bond representations.
                'bonds': (...),

                'InChI': 'The InChI of the molecule',
                'InChIKey': 'The InChIKey of the molecule',
            }

        For every :class:`.MoleculeKeyMaker` provided to `key_makers`,
        a new key will be added to the JSON representation, with its name
        given by :meth:`.MoleculeKeyMaker.get_key_name` and the value
        given by :meth:`.MoleculeKeyMaker.get_key`.
    """

    def __init__(
        self,
        key_makers: Iterable[MoleculeKeyMaker] = (InchiKey(),),
    ) -> None:
        """
        Parameters:
            key_makers (list[MoleculeKeyMaker]):
                Used to make the keys of molecules, which should be
                included in their JSON representations. Keys allow
                molecular data to reference itself when split across
                multiple JSONs.
        """
        self._jsonizer = MoleculeJsonizer(key_makers=())
        self._key_makers = tuple(key_makers)

    def to_json(self, molecule: ConstructedMolecule) -> dict:
        """
        Serialize `molecule`.

        Parameters:
            molecule:
                The constructed molecule to serialize.

        Returns:
            A JSON representation of `molecule`.
        """

        def get_keys(building_block: Molecule) -> dict:
            return {
                key_maker.get_key_name(): key_maker.get_key(building_block)
                for key_maker in self._key_makers
                if isinstance(key_maker, MoleculeKeyMaker)
            }

        building_block_indices: dict[Molecule | None, int | None] = {
            building_block: index
            for index, building_block in enumerate(
                molecule.get_building_blocks()
            )
        }
        building_block_indices[None] = None

        def atom_info_to_json(
            atom_info: AtomInfo,
        ) -> tuple[int | None, int | None, int | None]:
            building_block = atom_info.get_building_block()
            building_block_atom = atom_info.get_building_block_atom()
            if building_block is not None and building_block_atom is not None:
                return (
                    building_block_indices[building_block],
                    atom_info.get_building_block_id(),
                    building_block_atom.get_id(),
                )
            return (
                None,
                None,
                None,
            )

        def bond_info_to_json(
            bond_info: BondInfo,
        ) -> tuple[int | None, int | None]:
            return (
                building_block_indices[bond_info.get_building_block()],
                bond_info.get_building_block_id(),
            )

        molecule_json = self._jsonizer.to_json(molecule)
        constructed_molecule_json = {
            "BB": tuple(
                map(
                    get_keys,
                    molecule.get_building_blocks(),
                )
            ),
            "aI": tuple(
                map(
                    atom_info_to_json,
                    molecule.get_atom_infos(),
                )
            ),
            "bI": tuple(
                map(
                    bond_info_to_json,
                    molecule.get_bond_infos(),
                )
            ),
            "nBB": tuple(
                map(
                    molecule.get_num_building_block,
                    molecule.get_building_blocks(),
                )
            ),
        }
        for key_maker in self._key_makers:
            key_name = key_maker.get_key_name()
            key = key_maker.get_key(molecule)
            molecule_json["molecule"][key_name] = key  # type: ignore
            molecule_json["matrix"][key_name] = key  # type: ignore
            constructed_molecule_json[key_name] = key  # type: ignore

        building_block_jsons = tuple(
            map(
                self._jsonizer.to_json,
                molecule.get_building_blocks(),
            )
        )

        def is_molecule_key_maker(key_maker):
            return isinstance(key_maker, MoleculeKeyMaker)

        for key_maker in filter(
            is_molecule_key_maker,
            self._key_makers,
        ):
            key_name = key_maker.get_key_name()
            for building_block, json in zip(
                molecule.get_building_blocks(),
                building_block_jsons,
            ):
                key = key_maker.get_key(building_block)
                json["molecule"][key_name] = key  # type: ignore
                json["matrix"][key_name] = key  # type: ignore

        return {
            "molecule": molecule_json["molecule"],
            "constructedMolecule": constructed_molecule_json,
            "matrix": molecule_json["matrix"],
            "buildingBlocks": building_block_jsons,
        }

    def __str__(self) -> str:
        return repr(self)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._key_makers!r})"


class ConstructedMoleculeDejsonizer:
    """
    Abstract base class for creating constructed molecules from JSONs.

    See Also:
        :class:`.MoleculeDejsonizer`

    Notes:
        You might notice that the public methods of this abstract base
        class are implemented. These are just default implementations,
        which can be safely ignored or overridden, when implementing
        subclasses. However, the default implementation can be used
        directly, if it suits your needs.
    """

    def __init__(self) -> None:
        self._dejsonizer = MoleculeDejsonizer()

    def from_json(self, json: dict) -> ConstructedMolecule:
        """
        Get a :class:`.ConstructedMolecule` from a JSON.

        Parameters:
            json:
                A JSON of the constructed molecule.

        Returns:
            The constructed molecule.
        """
        building_blocks = tuple(
            map(
                self._dejsonizer.from_json,
                json["buildingBlocks"],
            )
        )

        num_building_blocks = {
            building_block: num
            for building_block, num in zip(
                building_blocks,
                json["constructedMolecule"]["nBB"],
            )
        }
        atoms = tuple(
            to_atom(atom_id, atom_json)
            for atom_id, atom_json in enumerate(json["molecule"]["a"])
        )
        bonds = tuple(
            to_bond(atoms, bond_json) for bond_json in json["molecule"]["b"]
        )
        return ConstructedMolecule.init(
            atoms=atoms,
            bonds=bonds,
            position_matrix=np.array(json["matrix"]["m"]),
            atom_infos=tuple(
                to_atom_info(
                    building_blocks=building_blocks,
                    atom=atoms[atom_id],
                    json=atom_info_json,
                )
                for atom_id, atom_info_json in enumerate(
                    json["constructedMolecule"]["aI"]
                )
            ),
            bond_infos=tuple(
                to_bond_info(
                    building_blocks=building_blocks,
                    bond=bonds[bond_id],
                    json=bond_info_json,
                )
                for bond_id, bond_info_json in enumerate(
                    json["constructedMolecule"]["bI"]
                )
            ),
            num_building_blocks=num_building_blocks,
        )

    def __str__(self) -> str:
        return repr(self)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"



import itertools as it


def check_atoms(old_state, new_state, building_blocks):
    old_atoms = tuple(old_state.get_atoms())
    new_atoms = tuple(new_state.get_atoms())
    num_added = len(new_atoms) - len(old_atoms)
    assert num_added == sum(bb.get_num_atoms() for bb in building_blocks)
    expected_atoms = enumerate(
        it.chain(*(bb.get_atoms() for bb in building_blocks)),
        len(old_atoms),
    )
    for atom1, (id_, atom2) in zip(
        # Take just the newly added atoms.
        new_atoms[len(old_atoms) :],
        expected_atoms,
    ):
        is_equivalent_atom(atom1, atom2.with_id(id_))


def is_equivalent_atom(atom1, atom2):
    assert atom1.get_id() == atom2.get_id()
    assert atom1.__class__ is atom2.__class__
    assert atom1.get_charge() == atom2.get_charge()



import itertools
import typing
from collections.abc import Sequence

import stk

from .case_data import CaseData

T = typing.TypeVar("T", bound=stk.MoleculeRecord)


def test_select(case_data: CaseData) -> None:
    _test_select(
        selector=case_data.selector,
        population=case_data.population,
        selected=case_data.selected,
    )


def _test_select(
    selector: stk.Selector[T],
    population: dict[T, float],
    selected: Sequence[stk.Batch[T]],
) -> None:
    inchi = stk.Inchi()
    for batch1, batch2 in itertools.zip_longest(
        sorted(selector.select(population), key=get_inchis),
        sorted(selected, key=get_inchis),
    ):
        inchis1 = tuple(
            inchi.get_key(record.get_molecule()) for record in batch1
        )
        inchis2 = tuple(
            inchi.get_key(record.get_molecule()) for record in batch2
        )
        assert inchis1 == inchis2


def get_inchis(batch: stk.Batch[T]) -> tuple[str, ...]:
    inchi = stk.Inchi()
    return tuple(inchi.get_key(record.get_molecule()) for record in batch)



import numpy as np
import stk


class CaseData:
    """
    A test case.

    Attributes:

        molecule:
            The molecule to be tested.

        smiles:
            The smiles for the molecule.

        name:
            The name of the test case.

        position_matrix:
            The position matrix of the molecule.

    """

    position_matrix: np.ndarray

    def __init__(
        self,
        molecule: stk.Molecule,
        smiles: str,
        name: str,
    ) -> None:
        self.molecule = molecule
        self.smiles = smiles
        self.name = name
        self.position_matrix = molecule.get_position_matrix()

    def with_position_matrix(self, path):
        """
        Return a clone holding a position matrix.

        Parameters
        ----------
        path : :class:`str`
            The for to a dumped :class:`numpy.ndarray` position
            matrix.

        Returns
        -------
        :class:`.CaseData`
            The clone, holding the position matrix found in `path`.

        """

        clone = self.__class__.__new__(self.__class__)
        CaseData.__init__(clone, self.molecule, self.smiles, self.name)
        clone.position_matrix = np.load(path)
        return clone

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"{self.molecule}, {self.smiles!r}, {self.name!r})"
        )



from .reaction_result import ReactionResult


class Reaction:
    """
    An abstract base class for reactions.

    Reactions are used to add and remove atoms and bonds during
    :class:`.ConstructedMolecule` construction. Each subclass
    will implement a specific algorithm for doing this. Each
    :class:`.Reaction` instance should operate on a small set of
    directly relevant :class:`.FunctionalGroup` instances, and only
    modify the atoms and bonds of those instances. Normally, like 99%
    of the time, this should just be two functional groups, and you
    should ensure your topology graph only needs to react two
    functional groups at a time, if you can. However, :mod:`stk` does
    not actually care, and you can modify as many atoms and bonds as
    you want in any reaction.

    See Also
    --------
    :mod:`.reaction_factory`
        Used for automated creation of :class:`.Reaction` instances.
        Typically, :class:`.Reaction` instances are not created
        directly, but only through some kind of
        :class:`.ReactionFactory` instance.

    Notes
    -----
    You might notice that the public method of this abstract
    base class, :meth:`.get_result`, is implemented. This is purely for
    convenience when implementing subclasses. The implemented public
    method is simply a default implementation, which can be safely
    ignored or overridden, when implementing subclasses. The private
    methods are an implementation detail of this default
    implementation. However, they are not implemented. To use the
    default implementation of :meth:`.get_result`, the private methods
    it relies on need to be implemented.

    Examples
    --------
    *Subclass Implementation*

    The source code of the subclasses, listed in
    :mod:`~.reaction.reaction`, can serve as good examples.

    """

    def get_result(self):
        """
        Get the result of carrying out the reaction.

        Returns
        -------
        :class:`.ReactionResult`
            Holds the results of the reaction.

        """

        return ReactionResult(
            new_atoms=tuple(self._get_new_atoms()),
            new_bonds=tuple(self._get_new_bonds()),
            deleted_atoms=tuple(self._get_deleted_atoms()),
            deleted_bonds=tuple(self._get_deleted_bonds()),
        )

    def _get_new_atoms(self):
        """
        Yield the atoms added by the reaction.

        Yields
        ------
        :class:`tuple`
            A :class:`tuple` of the form ``(atom, position)``, holding
            an :class:`.Atom` added
            by the reaction and its position as a
            :class:`numpy.ndarray`. New atoms have a negative id, and
            will need to be assigned a new one when added to the
            :class:`.ConstructedMolecule`.

        """

        raise NotImplementedError()

    def _get_new_bonds(self):
        """
        Yield the bonds added by the reaction.

        Yields
        ------
        :class:`.Bond`
            A bond added by the reaction.

        """

        raise NotImplementedError()

    def _get_deleted_atoms(self):
        """
        Yield the atoms removed by the reaction.

        Yields
        ------
        :class:`.Atom`
            An atom deleted by the reaction.

        """

        raise NotImplementedError()

    def _get_deleted_bonds(self):
        """
        Yield the bonds removed by the reaction between existing atoms.

        Yields
        ------
        :class:`.Bond`
            An bond deleted by the reaction.

        """

        raise NotImplementedError()



"""
Linear Polymer Vertices
=======================

"""

import logging

from stk._internal.topology_graphs.vertex import Vertex

logger = logging.getLogger(__name__)


class LinearVertex(Vertex):
    """
    Represents a vertex in the middle of a linear polymer chain.

    """

    def __init__(self, id, position, flip):
        """
        Initialize a :class:`.LinearVertex` instance.

        Parameters
        ----------
        id : :class:`int`
            The id of the vertex.

        position : :class:`numpy.ndarray`
            The position of the vertex.

        flip : :class:`bool`
            If ``True`` any building block placed by the vertex will
            have its orientation along the chain flipped.

        """

        super().__init__(id, position)
        self._flip = flip

    def get_flip(self):
        """
        Return ``True`` if the vertex flips building blocks it places.

        Returns
        -------
        :class:`bool`
            ``True`` if the vertex flips building blocks it places.

        """

        return self._flip

    def clone(self):
        clone = super().clone()
        clone._flip = self._flip
        return clone

    def place_building_block(self, building_block, edges):
        assert building_block.get_num_functional_groups() == 2, (
            f"{building_block} needs to have exactly 2 functional "
            "groups but has "
            f"{building_block.get_num_functional_groups()}."
        )
        building_block = building_block.with_centroid(
            position=self._position,
            atom_ids=building_block.get_placer_ids(),
        )
        fg1, fg2 = building_block.get_functional_groups()
        fg1_position = building_block.get_centroid(
            atom_ids=fg1.get_placer_ids(),
        )
        fg2_position = building_block.get_centroid(
            atom_ids=fg2.get_placer_ids(),
        )
        return building_block.with_rotation_between_vectors(
            start=fg2_position - fg1_position,
            target=[-1 if self._flip else 1, 0, 0],
            origin=self._position,
        ).get_position_matrix()

    def map_functional_groups_to_edges(self, building_block, edges):
        fg1_id, fg2_id = self._sort_functional_groups(building_block)
        edge1_id, edge2_id = self._sort_edges(edges)
        return {
            fg1_id: edge1_id,
            fg2_id: edge2_id,
        }

    @staticmethod
    def _sort_functional_groups(building_block):
        fg1, fg2 = building_block.get_functional_groups()
        x1, y1, z1 = building_block.get_centroid(
            atom_ids=fg1.get_placer_ids(),
        )
        x2, y2, z2 = building_block.get_centroid(
            atom_ids=fg2.get_placer_ids(),
        )
        return (0, 1) if x1 < x2 else (1, 0)

    @staticmethod
    def _sort_edges(edges):
        edge1, edge2 = edges
        x1, y1, z1 = edge1.get_position()
        x2, y2, z2 = edge2.get_position()
        if x1 < x2:
            return edge1.get_id(), edge2.get_id()
        else:
            return edge2.get_id(), edge1.get_id()

    def __str__(self):
        return (
            f"Vertex(id={self._id}, "
            f"position={self._position.tolist()}, "
            f"flip={self._flip})"
        )


class TerminalVertex(LinearVertex):
    """
    Represents a vertex at the end of a polymer chain.

    Do not instantiate this class directly, use :class:`.HeadVertex`
    or :class:`.TailVertex` instead.

    """

    def place_building_block(self, building_block, edges):
        if (
            building_block.get_num_functional_groups() != 1
            and building_block.get_num_placers() > 1
        ):
            return super().place_building_block(building_block, edges)

        building_block = building_block.with_centroid(
            position=self._position,
            atom_ids=building_block.get_placer_ids(),
        )
        fg, *_ = building_block.get_functional_groups()
        fg_centroid = building_block.get_centroid(
            atom_ids=fg.get_placer_ids(),
        )
        core_centroid = building_block.get_centroid(
            atom_ids=building_block.get_core_atom_ids(),
        )
        return building_block.with_rotation_between_vectors(
            start=fg_centroid - core_centroid,
            # _cap_direction is defined by a subclass.
            target=[self._cap_direction, 0, 0],
            origin=self._position,
        ).get_position_matrix()

    def map_functional_groups_to_edges(self, building_block, edges):
        if building_block.get_num_functional_groups() == 2:
            functional_groups = self._sort_functional_groups(
                building_block=building_block,
            )
            index = 1 if self._cap_direction == 1 else 0
            return {functional_groups[index]: edges[0].get_id()}

        elif building_block.get_num_functional_groups() == 1:
            return {0: edges[0].get_id()}

        else:
            raise ValueError(
                "The building block of a polymer "
                "must have 1 or 2 functional groups."
            )


class HeadVertex(TerminalVertex):
    """
    Represents a vertex at the head of a polymer chain.

    """

    # The direction to use if the building block placed on the
    # vertex only has 1 FunctionalGroup.
    _cap_direction = 1


class TailVertex(TerminalVertex):
    """
    Represents a vertex at the tail of a polymer chain.

    """

    # The direction to use if the building block placed on the
    # vertex only has 1 FunctionalGroup.
    _cap_direction = -1


class UnaligningVertex(LinearVertex):
    """
    Just places a building block, does not align.

    """

    def place_building_block(self, building_block, edges):
        return building_block.with_centroid(
            position=self._position,
            atom_ids=building_block.get_placer_ids(),
        ).get_position_matrix()

    def map_functional_groups_to_edges(self, building_block, edges):
        return {fg_id: edge.get_id() for fg_id, edge in enumerate(edges)}



import pytest
import stk


@pytest.fixture(
    params=[
        lambda bond: {},
        lambda bond: {bond.get_atom1().get_id(): stk.Be(33)},
        lambda bond: {bond.get_atom2().get_id(): stk.Bi(122)},
        lambda bond: {
            bond.get_atom1().get_id(): stk.K(4),
            bond.get_atom2().get_id(): stk.S(7),
        },
    ],
)
def get_atom_map(request):
    """
    Return a valid `atom_map` parameter for a bond.

    Parameters
    ----------
    bond : :class:`.Bond`
        The bond, for which the `atom_map` parameter needs to be
        created.

    Returns
    -------
    :class:`dict`
        A valid `atom_map` parameter for :meth:`.Bond.with_atoms`.

    """

    return request.param



import pytest
import stk

from .case_data import CaseData


@pytest.fixture(
    params=(
        lambda: CaseData(
            molecule=stk.BuildingBlock("NCCN"),
            bonds=(
                stk.Bond(stk.N(0), stk.C(1), 1),
                stk.Bond(stk.C(1), stk.C(2), 1),
                stk.Bond(stk.C(2), stk.N(3), 1),
                stk.Bond(stk.N(0), stk.H(4), 1),
                stk.Bond(stk.N(0), stk.H(5), 1),
                stk.Bond(stk.C(1), stk.H(6), 1),
                stk.Bond(stk.C(1), stk.H(7), 1),
                stk.Bond(stk.C(2), stk.H(8), 1),
                stk.Bond(stk.C(2), stk.H(9), 1),
                stk.Bond(stk.N(3), stk.H(10), 1),
                stk.Bond(stk.N(3), stk.H(11), 1),
            ),
        ),
    ),
)
def case_data(request) -> CaseData:
    return request.param()



def test_to_string(case_data):
    """
    Test writing of molecule to a string.

    Parameters
    ----------
    case_data : :class:`.CaseData`
        A test case.

    Returns
    -------
    None : :class:`NoneType`

    """

    _test_to_string(
        molecule=case_data.molecule,
        writer=case_data.writer,
        string=case_data.string,
    )


def _test_to_string(molecule, writer, string):
    """
    Test that the written string matches expected string.

    Parameters
    ----------
    molecule : :class:`.Molecule`
        Molecule to test.

    writer : :class:`.XyzWriter`
        The writer to test.

    string : :class:`str`
        The expected output string.

    Returns
    -------
    None : :class:`NoneType`

    """

    test_string = writer.to_string(molecule)

    assert test_string == string



import pathlib
import typing
from collections.abc import Iterable, Sequence

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

plt.switch_backend("agg")


class ProgressPlotter:
    """
    Plots how a property changes during an EA run.

    The produced plot will show the EA generations on the x axis and
    the min, mean and max values of an attribute on the y axis.

    Examples:

        *Plotting How Fitness Values Change Across Generations*

        .. testcode:: plotting-how-fitness-values-change-across-generations

            import stk

            # Initialize an EA somehow.
            ea = stk.EvolutionaryAlgorithm(
                initial_population=[
                    stk.MoleculeRecord(
                        topology_graph=stk.polymer.Linear(
                            building_blocks=[
                                stk.BuildingBlock(
                                    smiles='BrCCBr',
                                    functional_groups=stk.BromoFactory(),
                                ),
                            ],
                            repeating_unit='A',
                            num_repeating_units=i,
                        ),
                    )
                    for i in range(2, 22)
                ],
                fitness_calculator=stk.FitnessFunction(
                    fitness_function=lambda record:
                        record.get_molecule().get_num_atoms(),
                ),
                mutator=stk.RandomBuildingBlock(
                    building_blocks=[
                        stk.BuildingBlock('BrC[Si]CCBr', stk.BromoFactory()),
                        stk.BuildingBlock('BrCCCCCCCBr', stk.BromoFactory()),
                    ],
                    is_replaceable=lambda building_block: True
                ),
                crosser=stk.GeneticRecombination(
                    get_gene=lambda building_block: 0
                ),
                generation_selector=stk.Best(
                    num_batches=22,
                    duplicate_molecules=False,
                ),
                mutation_selector=stk.Roulette(
                    num_batches=5,
                    random_seed=10,
                ),
                crossover_selector=stk.Roulette(
                    num_batches=5,
                    batch_size=2,
                    random_seed=10,
                ),
                num_processes=1,
            )

            fitness_values = []
            for generation in ea.get_generations(10):
                fitness_values.append(
                    [
                        fitness_value.raw
                        for fitness_value
                        in generation.get_fitness_values().values()
                    ]
                )

            # Make the plotter which plots the fitness change across
            # generations.
            progress = stk.ProgressPlotter(
                property=fitness_values,
                y_label='Fitness'
            )
            progress.write('fitness_plot.png')

        .. testcode:: plotting-how-fitness-values-change-across-generations
            :hide:

            import os

            assert os.path.exists('fitness_plot.png')
            os.remove('fitness_plot.png')

        *Plotting How a Molecular Property Changes Across Generations*

        As an example, plotting how the number of atoms changes across
        generations

        .. testcode:: plotting-how-a-molecular-property-changes

            import stk

            # Initialize an EA somehow.
            ea = stk.EvolutionaryAlgorithm(
                initial_population=[
                    stk.MoleculeRecord(
                        topology_graph=stk.polymer.Linear(
                            building_blocks=[
                                stk.BuildingBlock(
                                    smiles='BrCCBr',
                                    functional_groups=stk.BromoFactory(),
                                ),
                            ],
                            repeating_unit='A',
                            num_repeating_units=i,
                        ),
                    )
                    for i in range(2, 22)
                ],
                fitness_calculator=stk.FitnessFunction(
                    fitness_function=lambda record:
                        record.get_molecule().get_num_atoms(),
                ),
                mutator=stk.RandomBuildingBlock(
                    building_blocks=[
                        stk.BuildingBlock('BrC[Si]CCBr', stk.BromoFactory()),
                        stk.BuildingBlock('BrCCCCCCCBr', stk.BromoFactory()),
                    ],
                    is_replaceable=lambda building_block: True
                ),
                crosser=stk.GeneticRecombination(
                    get_gene=lambda building_block: 0
                ),
                generation_selector=stk.Best(
                    num_batches=22,
                    duplicate_molecules=False,
                ),
                mutation_selector=stk.Roulette(
                    num_batches=5,
                    random_seed=10,
                ),
                crossover_selector=stk.Roulette(
                    num_batches=5,
                    batch_size=2,
                    random_seed=10,
                ),
                num_processes=1,
            )

            num_atoms = []
            for generation in ea.get_generations(10):
                num_atoms.append(
                    [
                        record.get_molecule().get_num_atoms()
                        for record in generation.get_molecule_records()
                    ]
                )

            # Make the plotter which plots the number of atoms across
            # generations.
            progress = stk.ProgressPlotter(
                property=num_atoms,
                y_label='Number of Atoms'
            )
            progress.write('number_of_atoms_plot.png')

        .. testcode:: plotting-how-a-molecular-property-changes
            :hide:

            import os

            assert os.path.exists('number_of_atoms_plot.png')
            os.remove('number_of_atoms_plot.png')

        *Excluding Molecules From the Plot*

        Sometimes, you want to ignore some molecules from the plot you
        make. For example, If the fitness calculation failed on a
        molecule, you not want to include in a plot of fitness.

        .. testcode:: excluding-molecules-from-the-plot

            import stk

            # Initialize an EA somehow.
            ea = stk.EvolutionaryAlgorithm(
                initial_population=[
                    stk.MoleculeRecord(
                        topology_graph=stk.polymer.Linear(
                            building_blocks=[
                                stk.BuildingBlock(
                                    smiles='BrCCBr',
                                    functional_groups=stk.BromoFactory(),
                                ),
                            ],
                            repeating_unit='A',
                            num_repeating_units=i,
                        ),
                    )
                    for i in range(2, 22)
                ],
                fitness_calculator=stk.FitnessFunction(
                    fitness_function=lambda record:
                        record.get_molecule().get_num_atoms(),
                ),
                mutator=stk.RandomBuildingBlock(
                    building_blocks=[
                        stk.BuildingBlock('BrC[Si]CCBr', stk.BromoFactory()),
                        stk.BuildingBlock('BrCCCCCCCBr', stk.BromoFactory()),
                    ],
                    is_replaceable=lambda building_block: True
                ),
                crosser=stk.GeneticRecombination(
                    get_gene=lambda building_block: 0
                ),
                generation_selector=stk.Best(
                    num_batches=22,
                    duplicate_molecules=False,
                ),
                mutation_selector=stk.Roulette(
                    num_batches=5,
                    random_seed=10,
                ),
                crossover_selector=stk.Roulette(
                    num_batches=5,
                    batch_size=2,
                    random_seed=10,
                ),
                num_processes=1,
            )

            fitness_values = []
            for generation in ea.get_generations(10):
                fitness_values.append(
                    [
                        fitness_value.normalized
                        for fitness_value
                        in generation.get_fitness_values().values()
                        if fitness_value.raw is not None
                    ]
                )

            # Make the plotter which plots the fitness change across
            # generations.
            progress = stk.ProgressPlotter(
                property=fitness_values,
                y_label='Fitness',
            )
            progress.write('fitness_plot.png')

        .. testcode:: excluding-molecules-from-the-plot
            :hide:

            import os

            assert os.path.exists('fitness_plot.png')
            os.remove('fitness_plot.png')
    """

    def __init__(
        self,
        property: Iterable[Sequence[float]],
        y_label: str,
    ) -> None:
        """
        Parameters:
            property (list[list[float]]):
                The generations of the EA, which are plotted.
            y_label:
                The y label for the produced graph.
        """
        self._property = tuple(property)
        self._y_label = y_label
        self._filter = filter
        self._plot_data = self._get_plot_data()

    def _get_plot_data(self) -> pd.DataFrame:
        self._num_generations = 0
        data = []
        for id_, property in enumerate(self._property):
            self._num_generations += 1

            # If there are no values after filtering, don't plot
            # anything for the generation.
            if not property:
                continue

            data.append(
                pd.DataFrame(
                    data={
                        "Generation": [id_, id_, id_],
                        self._y_label: [
                            max(property),
                            np.mean(property),
                            min(property),
                        ],
                        "Type": ["Max", "Mean", "Min"],
                    },
                    index=["Generation", "Generation", "Generation"],
                ),
            )
        return pd.concat(data, ignore_index=True)

    def get_plot_data(self) -> pd.DataFrame:
        """
        Get the plot data.

        Returns:
            A data frame holding the plot data.
        """
        return self._plot_data.copy()

    def write(self, path: pathlib.Path | str, dpi: int = 500) -> typing.Self:
        """
        Write a progress plot to a file.

        Parameters:

            path:
                The path into which the plot is written.

            dpi:
                The dpi of the image.

        Returns:
            ProgressPlotter: The plotter is returned.
        """
        sns.set(style="darkgrid")
        fig = plt.figure(figsize=[8, 4.5])
        palette = sns.color_palette("deep")

        # It's possible that all values were filtered out, and trying
        # to plot an empty dataframe would raise an exception.
        if len(self._plot_data) != 0:
            sns.scatterplot(
                x="Generation",
                y=self._y_label,
                hue="Type",
                palette={
                    "Max": palette[3],
                    "Min": palette[0],
                    "Mean": palette[2],
                },
                data=self._plot_data,
            )
        # Set the length of the axes to account for all generations,
        # as its possible the first or last ones were not included
        # due to being filtered out.
        plt.xlim(0, self._num_generations)

        plt.legend(bbox_to_anchor=(1.15, 1), prop={"size": 9})
        plt.tight_layout()
        fig.savefig(path, dpi=dpi)
        plt.close("all")
        return self



import numpy as np
import stk

from .building_block import is_clone_building_block
from .constructed_molecule import is_clone_constructed_molecule
from .molecule import is_clone_molecule


def normalize_ids(molecule, ids):
    if ids is None:
        return range(molecule.get_num_atoms())
    elif isinstance(ids, int):
        return (ids,)
    else:
        return ids


def has_same_structure(molecule1, molecule2):
    assert np.all(
        np.equal(
            molecule1.get_position_matrix(),
            molecule2.get_position_matrix(),
        )
    )


def get_displacement_vector(molecule, start_atom, end_atom):
    """
    Get the displacement vector between `start_atom` and `end_atom`.

    """

    position1, position2 = molecule.get_atomic_positions(
        (start_atom, end_atom)
    )
    return position2 - position1


def get_num_atom_ids(molecule, get_atom_ids):
    atom_ids = get_atom_ids(molecule)
    if atom_ids is None:
        return molecule.get_num_atoms()
    else:
        return len(tuple(atom_ids))


def is_clone(molecule1, molecule2):
    is_building_block1 = isinstance(molecule1, stk.BuildingBlock)
    is_building_block2 = isinstance(molecule2, stk.BuildingBlock)
    if is_building_block1 and is_building_block2:
        is_clone_building_block(molecule1, molecule2)
        return

    is_constructed_molecule1 = isinstance(
        molecule1,
        stk.ConstructedMolecule,
    )
    is_constructed_molecule2 = isinstance(
        molecule2,
        stk.ConstructedMolecule,
    )
    if is_constructed_molecule1 and is_constructed_molecule2:
        is_clone_constructed_molecule(molecule1, molecule2)
        return

    is_clone_molecule(molecule1, molecule2)



import pytest
from pytest_lazyfixture import lazy_fixture

# All fixtures must be visible for lazy_fixture() call.
from .octahedral import *  # noqa
from .paddlewheel import *  # noqa
from .porphyrin import *  # noqa
from .square_planar import *  # noqa


@pytest.fixture(
    params=(
        lazy_fixture("metal_complex_octahedral"),
        lazy_fixture("metal_complex_octahedral_lambda"),
        lazy_fixture("metal_complex_octahedral_delta"),
        lazy_fixture("metal_complex_porphyrin"),
        lazy_fixture("metal_complex_paddlewheel"),
        lazy_fixture("metal_complex_square_planar"),
        lazy_fixture("metal_complex_bidentate_square_planar"),
        lazy_fixture("metal_complex_cis_protected_square_planar"),
    ),
)
def metal_complex(request):
    return request.param



import spindry as spd

from stk._internal.construction_state.construction_state import (
    ConstructionState,
)

from .optimizer import Optimizer


class Spinner(Optimizer):
    """
    Performs Monte Carlo optimisation of host-guest complexes.

    Examples:

        *Structure Optimization*

        Using :class:`.Spinner` will lead to
        :class:`.ConstructedMolecule` structures with better host-guest
        structures. Especially useful for multiple-guest systems and
        removing overlap.

        .. testcode:: structure-optimization

            import stk

            bb1 = stk.BuildingBlock(
                smiles='NCCN',
                functional_groups=[stk.PrimaryAminoFactory()],
            )
            bb2 = stk.BuildingBlock(
                smiles='O=CC(C=O)C=O',
                functional_groups=[stk.AldehydeFactory()],
            )
            guest1 = stk.host_guest.Guest(
                building_block=stk.BuildingBlock('c1ccccc1'),
            )
            guest2 = stk.host_guest.Guest(
                building_block=stk.BuildingBlock('C1CCCCC1'),
            )
            cage = stk.ConstructedMolecule(
                topology_graph=stk.cage.FourPlusSix(
                    building_blocks=(bb1, bb2),
                    optimizer=stk.MCHammer(),
                ),
            )

            complex = stk.ConstructedMolecule(
                topology_graph=stk.host_guest.Complex(
                    host=stk.BuildingBlock.init_from_molecule(cage),
                    guests=(guest1, guest2),
                    optimizer=stk.Spinner(),
                ),
            )

        Optimisation with :mod:`stk` simply collects the final position
        matrix. The optimisation's trajectory can be output using the
        :mod:`SpinDry` implementation if required by the user
        [#spindry]_. This code is entirely nonphysical and is,
        therefore, completely general to any chemistry.

    References:

        .. [#spindry] https://github.com/andrewtarzia/SpinDry

    """

    def __init__(
        self,
        step_size: float = 1.5,
        rotation_step_size: float = 5.0,
        num_conformers: int = 50,
        max_attempts: int = 1000,
        nonbond_epsilon: float = 5.0,
        beta: float = 2.0,
        random_seed: int = 1000,
    ) -> None:
        """
        Initialize an instance of :class:`.Spinner`.

        Parameters:
            step_size: The relative size of the step to take during
                step.

            rotation_step_size: The relative size of the rotation to
                take during step.

            num_conformers: Number of conformers to extract.

            max_attempts: Maximum number of MC moves to try to generate
                conformers.

            nonbond_epsilon: Value of epsilon used in the nonbonded
                potential in MC moves. Determines strength of the
                nonbonded potential.

            beta: Value of beta used in the in MC moves. Beta takes the
                place of the inverse boltzmann temperature.

            random_seed: Random seed to use for MC algorithm.

        """

        self._optimizer = spd.Spinner(
            step_size=step_size,
            rotation_step_size=rotation_step_size,
            num_conformers=num_conformers,
            max_attempts=max_attempts,
            potential_function=spd.SpdPotential(
                nonbond_epsilon=nonbond_epsilon,
            ),
            beta=beta,
            random_seed=random_seed,
        )

    def optimize(self, state: ConstructionState) -> ConstructionState:
        supramolecule = spd.SupraMolecule(
            atoms=(
                spd.Atom(
                    id=atom.get_id(),
                    element_string=atom.__class__.__name__,
                )
                for atom in state.get_atoms()
            ),
            bonds=(
                spd.Bond(
                    id=i,
                    atom_ids=(
                        bond.get_atom1().get_id(),
                        bond.get_atom2().get_id(),
                    ),
                )
                for i, bond in enumerate(state.get_bonds())
            ),
            position_matrix=state.get_position_matrix(),
        )

        conformer = self._optimizer.get_final_conformer(supramolecule)
        return state.with_position_matrix(
            position_matrix=conformer.get_position_matrix(),
        )



from stk._internal.topology_graphs.cage.cage import Cage
from stk._internal.topology_graphs.cage.eight_plus_sixteen import (
    EightPlusSixteen,
)
from stk._internal.topology_graphs.cage.eight_plus_twelve import (
    EightPlusTwelve,
)
from stk._internal.topology_graphs.cage.five_plus_ten import FivePlusTen
from stk._internal.topology_graphs.cage.four_plus_eight import FourPlusEight
from stk._internal.topology_graphs.cage.four_plus_four import FourPlusFour
from stk._internal.topology_graphs.cage.four_plus_six import FourPlusSix
from stk._internal.topology_graphs.cage.four_plus_six_2 import FourPlusSix2
from stk._internal.topology_graphs.cage.m2l4_lantern import M2L4Lantern
from stk._internal.topology_graphs.cage.m3l3_triangle import M3L3Triangle
from stk._internal.topology_graphs.cage.m3l6 import M3L6
from stk._internal.topology_graphs.cage.m4l4_square import M4L4Square
from stk._internal.topology_graphs.cage.m4l4_tetrahedron import M4L4Tetrahedron
from stk._internal.topology_graphs.cage.m4l6_tetrahedron import M4L6Tetrahedron
from stk._internal.topology_graphs.cage.m4l6_tetrahedron_spacer import (
    M4L6TetrahedronSpacer,
)
from stk._internal.topology_graphs.cage.m4l8 import M4L8
from stk._internal.topology_graphs.cage.m6l2l3_prism import M6L2L3Prism
from stk._internal.topology_graphs.cage.m6l12_cube import M6L12Cube
from stk._internal.topology_graphs.cage.m8l6_cube import M8L6Cube
from stk._internal.topology_graphs.cage.m12l24 import M12L24
from stk._internal.topology_graphs.cage.m24l48 import M24L48
from stk._internal.topology_graphs.cage.one_plus_one import OnePlusOne
from stk._internal.topology_graphs.cage.six_plus_eight import SixPlusEight
from stk._internal.topology_graphs.cage.six_plus_nine import SixPlusNine
from stk._internal.topology_graphs.cage.six_plus_twelve import SixPlusTwelve
from stk._internal.topology_graphs.cage.ten_plus_twenty import TenPlusTwenty
from stk._internal.topology_graphs.cage.three_plus_six import ThreePlusSix
from stk._internal.topology_graphs.cage.twelve_plus_thirty import (
    TwelvePlusThirty,
)
from stk._internal.topology_graphs.cage.twenty_plus_thirty import (
    TwentyPlusThirty,
)
from stk._internal.topology_graphs.cage.two_plus_four import TwoPlusFour
from stk._internal.topology_graphs.cage.two_plus_three import TwoPlusThree
from stk._internal.topology_graphs.cage.two_plus_two import TwoPlusTwo
from stk._internal.topology_graphs.cage.vertices import (
    AngledVertex,
    LinearVertex,
    NonLinearVertex,
    UnaligningVertex,
)

__all__ = [
    "Cage",
    "M2L4Lantern",
    "M3L3Triangle",
    "M4L4Square",
    "M4L4Tetrahedron",
    "M4L6Tetrahedron",
    "M4L6TetrahedronSpacer",
    "M6L2L3Prism",
    "M6L12Cube",
    "M8L6Cube",
    "M12L24",
    "M24L48",
    "M4L8",
    "M3L6",
    "SixPlusEight",
    "OnePlusOne",
    "FourPlusFour",
    "TwoPlusTwo",
    "EightPlusSixteen",
    "FivePlusTen",
    "SixPlusTwelve",
    "ThreePlusSix",
    "FourPlusSix",
    "FourPlusEight",
    "TwoPlusThree",
    "TenPlusTwenty",
    "TwoPlusFour",
    "SixPlusNine",
    "FourPlusSix2",
    "TwelvePlusThirty",
    "EightPlusTwelve",
    "TwentyPlusThirty",
    "LinearVertex",
    "NonLinearVertex",
    "UnaligningVertex",
    "AngledVertex",
]



import numpy as np
import pytest
import stk

from ..case_data import CaseData


@pytest.fixture(
    scope="session",
    params=(
        lambda: CaseData(
            fitness_calculator=stk.PropertyVector(
                property_functions=(
                    stk.Molecule.get_num_atoms,
                    stk.Molecule.get_num_bonds,
                    stk.Molecule.get_maximum_diameter,
                ),
            ),
            molecule=stk.BuildingBlock("BrCCBr").with_position_matrix(
                position_matrix=np.array(
                    [
                        [0, 0, 0],
                        [10, 0, 0],
                        [0, 0, 0],
                        [0, 0, 0],
                        [0, 0, 0],
                        [0, 0, 0],
                        [0, 0, 0],
                        [0, 0, 0],
                    ],
                    dtype=np.float64,
                ),
            ),
            fitness_value=(8, 7, 10),
        ),
    ),
)
def property_vector(request) -> CaseData:
    return request.param()



import numpy as np
import stk

from ...utilities import (
    DatabaseEntry,
    DatabaseState,
    assert_database_state,
)
from .utilities import get_database_state


def to_hashable(json):
    json = dict(json)
    return {
        "m": tuple(tuple(row) for row in json.pop("m")),
        **json,
    }


def test_update_1(mongo_client):
    """
    Test that existing entries are updated.

    """

    database_name = "_test_update_1"
    mongo_client.drop_database(database_name)

    database = stk.MoleculeMongoDb(
        mongo_client=mongo_client,
        database=database_name,
        put_lru_cache_size=0,
        get_lru_cache_size=0,
    )
    jsonizer = stk.MoleculeJsonizer()

    molecule = stk.BuildingBlock("CCC").with_canonical_atom_ordering()
    json = jsonizer.to_json(molecule)

    database.put(molecule)
    assert_database_state(
        state1=get_database_state(database),
        state2=DatabaseState(
            {
                DatabaseEntry(**json["molecule"]): 1,
                DatabaseEntry(**to_hashable(json["matrix"])): 1,
            }
        ),
    )

    molecule2 = molecule.with_position_matrix(
        position_matrix=np.zeros((molecule.get_num_atoms(), 3)),
    )
    json2 = jsonizer.to_json(molecule2)

    database.put(molecule2)
    assert_database_state(
        state1=get_database_state(database),
        state2=DatabaseState(
            {
                # Molecule JSON should be unchanged.
                DatabaseEntry(**json["molecule"]): 1,
                # Position matrix should be updated.
                DatabaseEntry(**to_hashable(json2["matrix"])): 1,
            }
        ),
    )


def test_update_2(mongo_client):
    """
    Test that existing entries are updated.

    In this test, you first create two separate entries, using
    different molecule keys. You then update both at the same time,
    with a database which uses both molecule keys.

    """

    database_name = "_test_update_2"
    mongo_client.drop_database(database_name)

    jsonizer1 = stk.MoleculeJsonizer()
    database1 = stk.MoleculeMongoDb(
        mongo_client=mongo_client,
        database=database_name,
        put_lru_cache_size=0,
        get_lru_cache_size=0,
        jsonizer=jsonizer1,
    )

    jsonizer2 = stk.MoleculeJsonizer(
        key_makers=(stk.Smiles(),),
    )
    database2 = stk.MoleculeMongoDb(
        mongo_client=mongo_client,
        database=database_name,
        put_lru_cache_size=0,
        get_lru_cache_size=0,
        jsonizer=jsonizer2,
    )

    jsonizer3 = stk.MoleculeJsonizer(
        key_makers=(
            stk.InchiKey(),
            stk.Smiles(),
        ),
    )
    database3 = stk.MoleculeMongoDb(
        mongo_client=mongo_client,
        database=database_name,
        put_lru_cache_size=0,
        get_lru_cache_size=0,
        jsonizer=jsonizer3,
    )

    molecule1 = stk.BuildingBlock("CCC").with_canonical_atom_ordering()
    json1 = jsonizer1.to_json(molecule1)

    database1.put(molecule1)
    assert_database_state(
        state1=get_database_state(database1),
        state2=DatabaseState(
            {
                DatabaseEntry(**json1["molecule"]): 1,
                DatabaseEntry(**to_hashable(json1["matrix"])): 1,
            }
        ),
    )

    # Should add another entry, as a different key maker is used.
    molecule2 = molecule1.with_position_matrix(
        position_matrix=np.zeros((molecule1.get_num_atoms(), 3)),
    )
    json2 = jsonizer2.to_json(molecule2)

    database2.put(molecule2)
    assert_database_state(
        state1=get_database_state(database1),
        state2=DatabaseState(
            {
                DatabaseEntry(**json1["molecule"]): 1,
                DatabaseEntry(**to_hashable(json1["matrix"])): 1,
                DatabaseEntry(**json2["molecule"]): 1,
                DatabaseEntry(**to_hashable(json2["matrix"])): 1,
            }
        ),
    )

    # Should update both entries as both key makers are used.
    molecule3 = molecule1.with_position_matrix(
        position_matrix=np.ones((molecule1.get_num_atoms(), 3)),
    )
    json3 = jsonizer3.to_json(molecule3)

    database3.put(molecule3)
    assert_database_state(
        state1=get_database_state(database1),
        state2=DatabaseState(
            {
                DatabaseEntry(**json3["molecule"]): 2,
                DatabaseEntry(**to_hashable(json3["matrix"])): 2,
            }
        ),
    )


def test_update_3(mongo_client):
    """
    Test that existing entries are updated.

    In this test, you first create one entry with two keys. Then
    update the entry with databases, each using 1 different key.
    No duplicate entries should be made in the database this way.

    """

    database_name = "_test_update_3"
    mongo_client.drop_database(database_name)

    jsonizer1 = stk.MoleculeJsonizer(
        key_makers=(
            stk.InchiKey(),
            stk.Smiles(),
        ),
    )
    database1 = stk.MoleculeMongoDb(
        mongo_client=mongo_client,
        database=database_name,
        put_lru_cache_size=0,
        get_lru_cache_size=0,
        jsonizer=jsonizer1,
    )

    jsonizer2 = stk.MoleculeJsonizer()
    database2 = stk.MoleculeMongoDb(
        mongo_client=mongo_client,
        database=database_name,
        put_lru_cache_size=0,
        get_lru_cache_size=0,
        jsonizer=jsonizer2,
    )

    jsonizer3 = stk.MoleculeJsonizer(
        key_makers=(stk.Smiles(),),
    )
    database3 = stk.MoleculeMongoDb(
        mongo_client=mongo_client,
        database=database_name,
        put_lru_cache_size=0,
        get_lru_cache_size=0,
        jsonizer=jsonizer3,
    )

    molecule1 = stk.BuildingBlock("CCC").with_canonical_atom_ordering()
    json1 = jsonizer1.to_json(molecule1)

    database1.put(molecule1)
    assert_database_state(
        state1=get_database_state(database1),
        state2=DatabaseState(
            {
                DatabaseEntry(**json1["molecule"]): 1,
                DatabaseEntry(**to_hashable(json1["matrix"])): 1,
            }
        ),
    )

    molecule2 = molecule1.with_position_matrix(
        position_matrix=np.zeros((molecule1.get_num_atoms(), 3)),
    )
    json2 = jsonizer2.to_json(molecule2)
    json2["matrix"] = dict(json1["matrix"])
    json2["matrix"]["m"] = jsonizer2.to_json(molecule2)["matrix"]["m"]

    database2.put(molecule2)
    assert_database_state(
        state1=get_database_state(database1),
        state2=DatabaseState(
            {
                DatabaseEntry(**json1["molecule"]): 1,
                DatabaseEntry(**to_hashable(json2["matrix"])): 1,
            }
        ),
    )

    molecule3 = molecule1.with_position_matrix(
        position_matrix=np.zeros((molecule1.get_num_atoms(), 3)),
    )
    json3 = jsonizer3.to_json(molecule3)
    json3["matrix"] = dict(json1["matrix"])
    json3["matrix"]["m"] = jsonizer3.to_json(molecule3)["matrix"]["m"]

    database3.put(molecule3)
    assert_database_state(
        state1=get_database_state(database1),
        state2=DatabaseState(
            {
                DatabaseEntry(**json1["molecule"]): 1,
                DatabaseEntry(**to_hashable(json3["matrix"])): 1,
            }
        ),
    )



"""
Key Maker Utilities
===================

"""

from __future__ import annotations

import rdkit.Chem.AllChem as rdkit

from stk._internal.molecule import Molecule


def get_inchi(
    molecule: Molecule,
) -> str:
    """
    Get the InChI of `molecule`.

    Parameters:

        molecule:
            The molecule whose InChI is required.

    Returns:

        The InChI.

    Raises:

        :class:`ValueError`
            If the InChI of `molecule` cannot be generated.

    """

    inchi = rdkit.MolToInchi(molecule.to_rdkit_mol())
    if inchi:
        return inchi
    raise ValueError("The InChI of {molecule} was empty.")


def get_inchi_key(
    molecule: Molecule,
) -> str:
    """
    Get the InChIKey of `molecule`.

    Parameters:

        molecule:
            The molecule whose InChIKey is needed.

    Returns:

        The InChIKey.

    Raises:

        :class:`ValueError`
            If the InChIKey of `molecule` cannot be generated.

    """

    key = rdkit.MolToInchiKey(molecule.to_rdkit_mol())
    if not key:
        raise ValueError(f"InChIKey of {molecule} is empty.")
    return key


def get_smiles(molecule: Molecule) -> str:
    """
    Get the RDKit canonical, isomeric SMILES of `molecule`.

    Parameters:

        molecule:
            The molecule whose SMILES is required.

    Returns:

        The SMILES.

    """

    rdkit_mol = molecule.with_canonical_atom_ordering().to_rdkit_mol()
    rdkit.SanitizeMol(rdkit_mol)
    rdkit.AssignStereochemistryFrom3D(rdkit_mol)
    rdkit_mol = rdkit.RemoveHs(rdkit_mol)

    return rdkit.MolToSmiles(
        mol=rdkit_mol,
        isomericSmiles=True,
        canonical=True,
    )



import pytest
import stk

from ....case_data import CaseData
from .utilities import get_centroid, get_closest_point, get_edges


@pytest.fixture
def head_1(position, flip, building_block_1):
    point1, point2 = points = (
        position + [-10, 0, 0],
        position + [10, 0, 0],
    )

    def get_centroid_point(building_block):
        return get_closest_point(
            points=points,
            point=get_centroid(building_block),
        )

    vertex = stk.polymer.HeadVertex(0, position, flip)
    return CaseData(
        vertex=vertex,
        edges=(tuple(get_edges(vertex))[1],),
        building_block=building_block_1,
        position=position,
        alignment_tests={get_centroid_point: point1},
        functional_group_edges={0: 1},
    )



import logging
import typing
from collections.abc import Callable
from typing import Any

import numpy as np

from .fitness_normalizer import FitnessNormalizer

logger = logging.getLogger(__name__)

T = typing.TypeVar("T")


class DivideByMean(FitnessNormalizer[T]):
    """
    Divides fitness values by the population mean.

    While this function can be used if the fitness value of each
    molecule in the population is a single
    number, it is most useful when the fitness value is a
    :class:`tuple` of numbers. In this case, it is necessary to somehow
    combine the numbers so that a single fitness value is produced.
    For example, take a fitness value which is the vector holding the
    properties ``[energy, diameter, num_atoms]``. For a given molecule
    these numbers may be something like ``[200,000, 12, 140]``. If we
    were to sum these numbers, the energy term would dominate the final
    fitness value. In order to combine these numbers we can divide them
    by the population averages. For example, if the average energy
    of molecules in the population is ``300,000`` the average diameter
    is ``10`` and the average number of atoms is ``70`` then the
    fitness vector would be scaled to ``[0.5, 1.2, 2]``. These
    numbers are now of a similar magnitude and can be summed to give a
    reasonable value. After division , each value represents how
    much better than the population average each property value is.
    In essence we have removed the units from each parameter.

    Examples:

        *Selectively Normalizing Fitness Values*

        Sometimes you do not want to normalize all the values in a
        population together. For example, if a failed fitness value
        calculation resulted in some records having a fitness value of
        ``None``, you would want to ignore these records from the
        normalization

        .. testcode:: selectively-normalizing-fitness-values

            import stk
            import numpy as np

            building_block = stk.BuildingBlock('BrCCBr', stk.BromoFactory())
            record1 = stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=[building_block],
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            )
            record2 = stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=[building_block],
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            )
            fitness_values = {
                record1: (1., 2., 3.),
                record2: None,
            }
            mean_scaler = stk.DivideByMean(
                # Only normalize values which are not None.
                filter=lambda fitness_values, record:
                    fitness_values[record] is not None
            )
            normalized_fitness_values = mean_scaler.normalize(
                fitness_values=fitness_values,
            )
            assert np.all(np.equal(
                normalized_fitness_values[record1],
                (1, 1, 1),
            ))
    """

    def __init__(
        self,
        filter: Callable[
            [dict[T, Any], T], bool
        ] = lambda fitness_values, record: True,
    ) -> None:
        """
        Parameters:
            filter:
                A function which returns ``True`` or ``False``. Only
                molecules which return ``True`` will have fitness values
                normalized. By default, all molecules will have fitness
                values normalized.
                The instance passed to the `fitness_values` argument of
                :meth:`.normalize` is passed as the first argument, while
                the second argument will be passed every
                :class:`.MoleculeRecord` in it, one at a time.
        """
        self._filter = filter

    def normalize(self, fitness_values: dict[T, Any]) -> dict[T, Any]:
        filtered = {
            record: fitness_value
            for record, fitness_value in fitness_values.items()
            if self._filter(fitness_values, record)
        }
        mean = np.mean(
            a=[fitness_values[record] for record in filtered],
            axis=0,
        )
        logger.debug(f"Means used: {mean}")

        return {
            record: np.divide(fitness_value, mean)
            if self._filter(fitness_values, record)
            else fitness_value
            for record, fitness_value in fitness_values.items()
        }



import pathlib

import pandas as pd
import stk

from .case_data import CaseData


def test_progress_plotter(tmp_path: pathlib.Path, case_data: CaseData) -> None:
    _test_progress_plotter(
        plotter=case_data.plotter,
        path=tmp_path / "plot.png",
        plot_data=case_data.plot_data,
    )


def _test_progress_plotter(
    plotter: stk.ProgressPlotter,
    path: pathlib.Path,
    plot_data: pd.DataFrame,
) -> None:
    plotter.write(path)
    assert plotter.get_plot_data().equals(plot_data)



import numpy as np


def test_get_cell(case_data):
    """
    Test :meth:`.Vertex.get_cell`.

    Parameters
    ----------
    case_data : :class:`.CaseData`
        A test case. Holds the vertex to test and the correct cell.

    Returns
    -------
    None : :class:`NoneType`

    """

    _test_get_cell(case_data.vertex, case_data.cell)


def _test_get_cell(vertex, cell):
    """
    Test :meth:`.Vertex.get_cell`.

    Parameters
    ----------
    vertex : :class:`.Vertex`
        The vertex to test.

    cell : :class:`numpy.ndarray`
        The correct cell.

    Returns
    -------
    None : :class:`NoneType`

    """

    assert np.all(np.equal(vertex.get_cell(), cell))
